import itertools
import logging
import os
import pickle
import sys
import time
from pathlib import Path
from typing import Optional, Sequence, Union

import anndata as ad
import numpy as np
import pandas as pd
import torch
from tqdm import trange

from popari._hierarchical_view import HierarchicalView, Hierarchy
from popari._popari_dataset import PopariDataset
from popari.io import load_anndata, merge_anndata, save_anndata, unmerge_anndata
from popari.util import convert_numpy_to_pytorch_sparse_coo, get_datetime


class Popari:
    r"""Popari optimization model.

    Models spatial biological data using the NMF-HMRF formulation. Supports multiple
    fields-of-view (FOVs) and differential analysis.

    Example of including math in docstring (for use later):
    :math:`||A||_F = [\sum_{i,j} abs(a_{i,j})^2]^{1/2}`

    Attributes:
        K: number of metagenes to learn
        replicate_names: names of spatial datasets
        datasets: list of input AnnData spatial datasets for Popari.
        dataset_path: path to AnnData merged dataset on disk. Ignored if ``datasets`` is specified.
        reloaded_hierarchy: data from previous hierarchical run of Popari.
        lambda_Sigma_x_inv: hyperparameter to balance importance of spatial information. Default: ``1e-4``
        pretrained: if set, attempts to load model state from input files. Default: ``False``
        initialization_method: algorithm to use for initializing metagenes and embeddings. Default: ``leiden``
        hierarchical_levels: number of hierarchical levels to use. Default: ``1`` (non-hierarchical mode)
        metagene_groups: defines a grouping of replicates for the metagene optimization. If
            ``metagene_mode == "shared"``, then one set of metagenes will be created for each group;
            if ``metagene_mode == "differential",  then all replicates will have their own set of metagenes,
            but each group will share an ``M_bar``.
        spatial_affinity_groups: defines a grouping of replicates for the spatial affinity optimization.
            If ``spatial_affinity_mode == "shared lookup"``, then one set of spatial_affinities will be created for each group;
            if ``spatial_affinity_mode == "differential lookup"``,  then all replicates will have their own set of spatial
            affinities, but each group will share a ``spatial_affinity_bar``.
        betas: weighting of each dataset during optimization. Defaults to equally weighting each dataset
        prior_x_modes: family of prior distribution for embeddings of each dataset
        M_constraint: constraint on columns of M. Default: ``simplex``
        sigma_yx_inv_mode: form of sigma_yx_inv parameter. Default: ``separate``
        torch_context: keyword args to use of PyTorch tensors during training.
        initial_context: keyword args to use during initialization of PyTorch tensors.
        metagene_mode: modality of metagene parameters. Default: ``shared``.

            =================  =====
            ``metagene_mode``  Option
            =================  =====
            ``shared``         A metagene set is shared between all replicates in a group.
            ``differential``   Each replicate learns its own metagene set.
            =================  =====

        spatial_affinity_mode: modality of spatial affinity parameters. Default: ``shared lookup``
        lambda_M: hyperparameter to constrain metagene deviation in differential case. Ignored if
            ``metagene_mode`` is ``shared``. Default: ``0.5``
        lambda_Sigma_bar: hyperparameter to constrain spatial affinity deviation in differential case. Ignored if
            ``spatial_affinity_mode`` is ``shared lookup``. Default: ``0.5``
        spatial_affinity_lr: learning rate for optimization of ``Sigma_x_inv``
        spatial_affinity_tol: convergence tolerance during optimization of ``Sigma_x_inv``
        spatial_affinity_constraint: method to ensure that spatial affinities lie within an appropriate range
        spatial_affinity_centering: if set, spatial affinities are zero-centered after every optimization step
        spatial_affinity_scaling: magnitude of spatial affinities during initial scaling. Default: ``10``
        spatial_affinity_regularization_power: exponent controlling penalization of spatial affinity magnitudes. Default: ``2``
        embedding_mini_iterations: number of mini-iterations to use during each iteration of embedding optimization. Default: ``1000``
        embedding_acceleration_trick: if set, use trick to accelerate convergence of embedding optimization. Default: ``True``
        embedding_step_size_multiplier: controls relative step size during embedding optimization. Default: ``1.0``
        binning_downsample_rate: ratio of number of spots at low resolution to high resolution when
            using hierarchical mode
        superresolution_lr: learning rate for optimization of ``X`` from low-res embeddings
        use_inplace_ops: if set, inplace PyTorch operations will be used to speed up computation
        random_state: seed for reproducibility of randomized computations. Default: ``0``
        verbose: level of verbosity to use during optimization. Default: ``0`` (no print statements)

    """

    def __init__(
        self,
        K: int,
        replicate_names: Optional[Sequence[str]] = None,
        datasets: Optional[Sequence[ad.AnnData]] = None,
        dataset_path: Optional[Union[str, Path]] = None,
        reloaded_hierarchy: Optional[dict] = None,
        lambda_Sigma_x_inv: float = 1e-4,
        pretrained: bool = False,
        initialization_method: str = "leiden",
        hierarchical_levels: int = 1,
        metagene_groups: Optional[dict] = None,
        spatial_affinity_groups: Optional[dict] = None,
        betas: Optional[Sequence[float]] = None,
        prior_x_modes: Optional[Sequence[str]] = None,
        M_constraint: str = "simplex",
        sigma_yx_inv_mode: str = "separate",
        torch_context: Optional[dict] = None,
        initial_context: Optional[dict] = None,
        metagene_mode: str = "shared",
        spatial_affinity_mode: str = "shared lookup",
        lambda_M: float = 0.5,
        lambda_Sigma_bar: float = 1e-3,
        spatial_affinity_lr: float = 1e-2,
        spatial_affinity_tol: float = 2e-3,
        spatial_affinity_constraint: Optional[str] = None,
        spatial_affinity_centering: bool = False,
        spatial_affinity_scaling: int = 10,
        spatial_affinity_regularization_power: int = 2,
        embedding_mini_iterations: int = 1000,
        embedding_acceleration_trick: bool = True,
        embedding_step_size_multiplier: float = 1.0,
        downsampling_method: str = "grid",
        binning_downsample_rate: float = 0.2,
        chunks: int = 2,
        superresolution_lr: float = 1e-1,
        use_inplace_ops: bool = True,
        random_state: int = 0,
        verbose: int = 0,
    ):

        self.use_inplace_ops = use_inplace_ops
        self.verbose = verbose

        if not any([datasets, dataset_path]):
            raise ValueError("At least one of `datasets`, `dataset_path` must be specified in the Popari constructor.")

        if K <= 1:
            raise ValueError("`K` must be an integer value greater than 1.")

        if not torch_context:
            torch_context = dict(device="cpu", dtype=torch.float32)

        if not initial_context:
            initial_context = dict(device="cpu", dtype=torch.float32)

        self.context = torch_context
        self.initial_context = initial_context

        self.random_state = random_state
        torch.manual_seed(self.random_state)
        np.random.seed(self.random_state)

        self.K = K
        self.lambda_Sigma_x_inv = lambda_Sigma_x_inv
        self.lambda_Sigma_bar = lambda_Sigma_bar
        self.spatial_affinity_lr = spatial_affinity_lr
        self.spatial_affinity_tol = spatial_affinity_tol
        self.spatial_affinity_constraint = spatial_affinity_constraint
        self.spatial_affinity_centering = spatial_affinity_centering
        self.spatial_affinity_scaling = spatial_affinity_scaling
        self.spatial_affinity_regularization_power = spatial_affinity_regularization_power
        self.M_constraint = M_constraint
        self.sigma_yx_inv_mode = sigma_yx_inv_mode
        self.spatial_affinity_mode = spatial_affinity_mode
        self.pretrained = pretrained

        self.metagene_mode = metagene_mode
        self.lambda_M = lambda_M
        self.metagene_groups = metagene_groups
        self.spatial_affinity_groups = spatial_affinity_groups

        self.embedding_step_size_multiplier = embedding_step_size_multiplier
        self.embedding_mini_iterations = embedding_mini_iterations
        self.embedding_acceleration_trick = embedding_acceleration_trick

        self.hierarchical_levels = hierarchical_levels
        self.reloaded_hierarchy = reloaded_hierarchy
        self.superresolution_lr = superresolution_lr
        self.downsampling_method = downsampling_method
        self.binning_downsample_rate = binning_downsample_rate
        self.chunks = chunks

        if dataset_path:
            self.load_dataset(dataset_path)
        elif datasets:
            self.load_anndata_datasets(datasets, replicate_names)

        if replicate_names is None:
            self.replicate_names = [dataset.name for dataset in self.datasets]
        else:
            self.replicate_names = [f"{replicate_name}" for replicate_name in replicate_names]

        self.parameter_optimizer_hyperparameters = {
            "lambda_Sigma_x_inv": self.lambda_Sigma_x_inv,
            "lambda_Sigma_bar": self.lambda_Sigma_bar,
            "spatial_affinity_lr": self.spatial_affinity_lr,
            "spatial_affinity_tol": self.spatial_affinity_tol,
            "spatial_affinity_constraint": self.spatial_affinity_constraint,
            "spatial_affinity_centering": self.spatial_affinity_centering,
            "spatial_affinity_scaling": self.spatial_affinity_scaling,
            "spatial_affinity_regularization_power": self.spatial_affinity_regularization_power,
            "M_constraint": self.M_constraint,
            "sigma_yx_inv_mode": self.sigma_yx_inv_mode,
            "spatial_affinity_mode": self.spatial_affinity_mode,
            "lambda_M": self.lambda_M,
            "metagene_mode": self.metagene_mode,
        }

        self.embedding_optimizer_hyperparameters = {
            "embedding_step_size_multiplier": embedding_step_size_multiplier,
            "embedding_mini_iterations": embedding_mini_iterations,
            "embedding_acceleration_trick": embedding_acceleration_trick,
        }

        self._initialize(betas=betas, prior_x_modes=prior_x_modes, method=initialization_method, pretrained=pretrained)

    def load_anndata_datasets(self, datasets: Sequence[ad.AnnData], replicate_names: Sequence[str]):
        """Load Popari data directly from AnnData objects.

        Args:
            datasets: spatial transcriptomics datasets in AnnData format (one for each FOV)
            replicate_names: names for all datasets/replicates

        """
        self.datasets = [
            PopariDataset(dataset, replicate_name) for dataset, replicate_name in zip(datasets, replicate_names)
        ]
        self.num_replicates = len(self.datasets)

    def load_dataset(self, dataset_path: Union[str, Path]):
        """Load dataset into Popari from saved .h5ad file.

        Args:
            dataset_path: path to input ST datasets, stored in .h5ad format

        """

        dataset_path = Path(dataset_path)

        datasets, replicate_names = load_anndata(dataset_path)
        self.load_anndata_datasets(datasets, replicate_names)

    def _initialize(
        self,
        pretrained=False,
        betas: Optional[Sequence[float]] = None,
        prior_x_modes: Optional[Sequence[str]] = None,
        method: str = "svd",
    ):
        """Initialize metagenes and hidden states.

        Args:
            betas: weighting of each dataset during optimization. Defaults to equally weighting each dataset
            prior_x_modes: family of prior distribution for embeddings of each dataset
            method: algorithm to use for initializing metagenes and embeddings. Default: SVD

        """

        hierarchical_view_kwargs = {
            "random_state": self.random_state,
            "K": self.K,
            "context": self.context,
            "initial_context": self.initial_context,
            "hierarchical_levels": self.hierarchical_levels,
            "betas": betas,
            "prior_x_modes": prior_x_modes,
            "use_inplace_ops": self.use_inplace_ops,
            "method": method,
            "pretrained": self.pretrained,
            "verbose": self.verbose,
            "metagene_groups": self.metagene_groups,
            "spatial_affinity_groups": self.spatial_affinity_groups,
            "superresolution_lr": self.superresolution_lr,
            "parameter_optimizer_hyperparameters": self.parameter_optimizer_hyperparameters,
            "embedding_optimizer_hyperparameters": self.embedding_optimizer_hyperparameters,
        }

        bin_assignment_kwargs = {}
        if self.downsampling_method == "grid":
            bin_assignment_kwargs["chunks"] = self.chunks
        elif self.downsampling_method == "partition":
            bin_assignment_kwargs["adjacency_list_key"] = "adjacency_list"

        self.base_view = HierarchicalView(self.datasets, level=0, **hierarchical_view_kwargs)

        if self.pretrained:
            self.hierarchy = Hierarchy.reconstruct(
                self.reloaded_hierarchy,
                **hierarchical_view_kwargs,
            )
        else:
            self.hierarchy = Hierarchy(
                downsampling_method=self.downsampling_method,
                base_view=self.base_view,
                **hierarchical_view_kwargs,
            )

            self.hierarchy.construct(
                levels=self.hierarchical_levels,
                downsample_rate=self.binning_downsample_rate,
                **bin_assignment_kwargs,
            )

        self.base_view = self.hierarchy[self.hierarchical_levels - 1]

        self.active_view = self.base_view

        self.datasets = self.active_view.datasets
        self.Ys = self.active_view.Ys
        self.betas = self.active_view.betas
        self.parameter_optimizer = self.active_view.parameter_optimizer
        self.embedding_optimizer = self.active_view.embedding_optimizer
        self.metagene_groups = self.active_view.metagene_groups
        self.metagene_tags = self.active_view.metagene_tags
        self.spatial_affinity_groups = self.active_view.spatial_affinity_groups
        self.spatial_affinity_tags = self.active_view.spatial_affinity_tags

        self.synchronize_datasets()

    def estimate_weights(self, use_neighbors: bool = True, synchronize: bool = True):
        """Update embeddings (latent states) for each replicate.

        Args:
            use_neighbors: If specified, weight updates will take into account neighboring
                interactions. Default: ``True``

        """
        if self.verbose:
            print(f"{get_datetime()} Updating latent states")
        self.embedding_optimizer.update_embeddings(use_neighbors=use_neighbors)

        if synchronize:
            self.synchronize_datasets()

    def estimate_parameters(
        self,
        update_spatial_affinities: bool = True,
        differentiate_spatial_affinities: bool = True,
        differentiate_metagenes: bool = True,
        simplex_projection_mode: bool = "exact",
        edge_subsample_rate: Optional[float] = None,
        synchronize: bool = True,
    ):
        """Update parameters for each replicate.

        Args:
            update_spatial_affinities: If specified, spatial affinities will be updated during
                this iteration. Default: ``True``
            edge_subsample_rate: Fraction of adjacency matrix edges that will be included in
                optimization of ``Sigma_x_inv``.

        """
        logging.info(f"{get_datetime()}Updating model parameters")

        if update_spatial_affinities:
            if self.verbose:
                print(f"{get_datetime()} Updating spatial affinities")
            self.parameter_optimizer.update_spatial_affinity(
                differentiate_spatial_affinities=differentiate_spatial_affinities,
                subsample_rate=edge_subsample_rate,
            )

        if self.verbose:
            print(f"{get_datetime()} Updating metagenes")

        self.parameter_optimizer.update_metagenes(
            differentiate_metagenes=differentiate_metagenes,
            simplex_projection_mode=simplex_projection_mode,
        )

        if self.verbose:
            print(f"{get_datetime()} Updating sigma_yx")

        self.parameter_optimizer.update_sigma_yx()

        if synchronize:
            self.synchronize_datasets()

    def superresolve(
        self,
        differentiate_spatial_affinities: bool = True,
        update_spatial_affinities: bool = True,
        edge_subsample_rate: Optional[float] = None,
        use_manual_gradients: bool = True,
        n_epochs: int = 10000,
        miniepochs: int = None,
        tol: float = 1e-5,
    ):
        """Superresolve embeddings in hierarchical case.

        Works in a cascading manner, by superresolving the embeddings from
        lowest to highest resolutions in order.

        """

        if miniepochs is None:
            miniepochs = min(100, n_epochs)

        effective_epochs = n_epochs // miniepochs + 1

        for level in range(self.hierarchical_levels - 2, -1, -1):
            if self.verbose:
                print(f"{get_datetime()} Superresolving level {level} embeddings")
            view = self.hierarchy[level]
            view._propagate_parameters()

            progress_bar = trange(effective_epochs, leave=True, disable=not self.verbose, miniters=10000)
            previous_losses = np.full(len(view.datasets), np.inf)
            for epoch in progress_bar:
                view.parameter_optimizer.update_sigma_yx()
                losses = view._superresolve_embeddings(
                    n_epochs=miniepochs,
                    tol=tol,
                    use_manual_gradients=use_manual_gradients,
                    verbose=self.verbose,
                )
                formatted_losses = [f"{loss:.1e}" for loss in losses]
                formatted_deltas = [f"{delta:.1e}" for delta in ((previous_losses - losses) / losses)]
                description = (
                    f"Updating weights hierarchically: loss = {formatted_losses} " f"%δloss = {formatted_deltas} "
                )
                previous_losses = losses
                progress_bar.set_description(description)

            pretrained_embeddings = [
                view.embedding_optimizer.embedding_state[dataset.name].clone() for dataset in view.datasets
            ]
            view.parameter_optimizer.spatial_affinity_state.initialize(pretrained_embeddings)

            if update_spatial_affinities:
                view.parameter_optimizer.update_spatial_affinity(
                    differentiate_spatial_affinities=differentiate_spatial_affinities,
                    subsample_rate=edge_subsample_rate,
                )

        self.synchronize_datasets()

    def nll(self, level: int = 0, use_spatial: bool = False):
        """Compute the nll for the current configuration of model parameters."""

        view = self.hierarchy[level]
        return view.nll(use_spatial=use_spatial)

    def set_superresolution_lr(self, new_lr: float, target_level: Optional[int] = None):
        """Change learning rate for superresolution optimization.

        Can be used to change learning rate for all hierarchical levels (by
        default) or just the learning rate for a certain resolution.

        """

        change_all = target_level == None
        for level in range(self.hierarchical_levels):
            if change_all or (level == target_level):
                self.hierarchy[level].superresolution_lr = new_lr

    def synchronize_datasets(self):
        """Synchronize datasets across all hierarchical levels."""

        self.base_view.synchronize_datasets()
        if self.hierarchical_levels > 1:
            for level in range(self.hierarchical_levels):
                self.hierarchy[level].synchronize_datasets()

    def save_results(self, dataset_path: str, ignore_raw_data: bool = True):
        """Save datasets and learned Popari parameters to disk.

        Args:
            dataset_path: where to save results. If results are non-hierarchical, file extension
                will automatically be changed to ``.h5ad``. Otherwise, if results are hierarchical,
                ``dataset_path`` will be interpreted as a path to a subfolder where separate
                ``.h5ad`` files will be stored for each hierarchical level.
            ignore_raw_data: if set, only learned parameters and embeddings will be saved; raw gene expression will be ignored.

        """

        dataset_path = Path(dataset_path)
        path_without_extension = dataset_path.parent / dataset_path.stem

        self.synchronize_datasets()

        if self.hierarchical_levels == 1:
            if self.verbose:
                print(f"{get_datetime()} Writing results to {path_without_extension}.h5ad")
            save_anndata(f"{path_without_extension}.h5ad", self.datasets, ignore_raw_data=ignore_raw_data)
        else:
            path_without_extension.mkdir(exist_ok=True)

            merged_datasets = []
            for level in range(self.hierarchical_levels):
                view = self.hierarchy[level]
                datasets = view.datasets
                if self.verbose:
                    print(
                        f"{get_datetime()} Writing hierarchical results to {path_without_extension / f'level_{level}.h5ad'}",
                    )

                save_anndata(path_without_extension / f"level_{level}.h5ad", datasets, ignore_raw_data=ignore_raw_data)

    def _reload_expression(self, raw_datasets: Sequence[PopariDataset]):
        """Can be used to recover expression values for training model if saved
        with `ignore_raw_data=True`"""
        high_resolution_view = self.hierarchy[0]
        for index, (raw_dataset, dataset) in enumerate(zip(raw_datasets, high_resolution_view.datasets)):
            dataset.X = raw_dataset.X.copy()
            num_cells, _ = dataset.shape

            Y = convert_numpy_to_pytorch_sparse_coo(dataset.X, self.context)
            Y *= (self.K * 1) / (Y.sum() / num_cells)
            high_resolution_view.Ys[index] = Y

        high_resolution_view.parameter_optimizer.update_sigma_yx()  # This is necessary, since `sigma_yx` isn't loaded correctly from disk

        for level in range(self.hierarchical_levels - 1):
            view = self.hierarchy[level]
            low_res_view = self.hierarchy[level + 1]
            for index, (dataset, binned_dataset, previous_Y) in enumerate(
                zip(view.datasets, low_res_view.datasets, view.Ys),
            ):
                bin_assignments = binned_dataset.obsm[f"bin_assignments_{binned_dataset.name}"]
                binned_expression = bin_assignments @ dataset.X
                binned_dataset.X = binned_expression

                bin_assignments_tensor = convert_numpy_to_pytorch_sparse_coo(
                    bin_assignments,
                    context=self.initial_context,
                )

                binned_Y = bin_assignments_tensor @ previous_Y
                low_res_view.Ys[index] = binned_Y

            low_res_view.parameter_optimizer.update_sigma_yx()


class SpiceMix(Popari):
    """Wrapper to produce SpiceMix hyperparameter configuration."""

    def __init__(self, **spicemix_hyperparameters):
        spatial_affinity_mode = "shared lookup"
        if "spatial_affinity_mode" in spicemix_hyperparameters:
            spicemix_hyperparameters.pop("spatial_affinity_mode")

        super().__init__(spatial_affinity_mode="shared lookup", **spicemix_hyperparameters)


def load_trained_model(
    dataset_path: Union[str, Path],
    context=dict(device="cpu", dtype=torch.float64),
    **popari_kwargs,
):
    """Load trained Popari model for downstream analysis.

    Args:
        dataset_path: location of Popari results, stored as a .h5ad file.

    """

    # TODO: change this so that replicate_names can rename the datasets in the saved file...?

    dataset_path = Path(dataset_path)
    path_without_extension = dataset_path.parent / dataset_path.stem

    datasets = reloaded_hierarchy = hierarchical_levels = None
    reloaded_hierarchy = {}
    if Path(f"{path_without_extension}.h5ad").exists():
        level = 0
        merged_dataset = ad.read_h5ad(dataset_path)
        datasets, replicate_names = unmerge_anndata(merged_dataset)
        reloaded_hierarchy[level] = datasets

        popari_kwargs["hierarchical_levels"] = 1

    elif path_without_extension.is_dir():
        for level_path in path_without_extension.iterdir():
            path_parts = level_path.stem.split("_")
            if len(path_parts) != 2:
                continue

            level = int(path_parts[-1])
            merged_dataset = ad.read_h5ad(level_path)
            datasets, replicate_names = unmerge_anndata(merged_dataset)
            reloaded_hierarchy[level] = datasets

        popari_kwargs["hierarchical_levels"] = len(reloaded_hierarchy)
    else:
        raise FileNotFoundError(f"No Popari model saved at {path_without_extension}.")

    datasets = reloaded_hierarchy[0]
    replicate_names = [dataset.name for dataset in datasets]

    return load_pretrained(
        datasets,
        replicate_names,
        reloaded_hierarchy=reloaded_hierarchy,
        context=context,
        **popari_kwargs,
    )


def load_pretrained(
    datasets: Sequence[PopariDataset],
    replicate_names: Sequence[str] = None,
    context=dict(device="cpu", dtype=torch.float64),
    reloaded_hierarchy: Optional[dict] = None,
    **popari_kwargs,
):
    """Load pretrained Popari model from in-memory datasets."""
    first_dataset = datasets[0]
    saved_hyperparameters = first_dataset.uns["popari_hyperparameters"]

    metagene_groups = saved_hyperparameters["metagene_groups"]
    for group in metagene_groups:
        metagene_groups[group] = list(metagene_groups[group])

    spatial_affinity_groups = saved_hyperparameters["spatial_affinity_groups"]
    for group in spatial_affinity_groups:
        spatial_affinity_groups[group] = list(spatial_affinity_groups[group])

    new_kwargs = saved_hyperparameters.copy()
    for keyword in popari_kwargs:
        new_kwargs[keyword] = popari_kwargs[keyword]

    for noninitial_hyperparameter in ["prior_x", "metagene_tags", "spatial_affinity_tags"]:
        new_kwargs.pop(noninitial_hyperparameter)

    # metagene_mode = saved_hyperparameters["metagene_mode"]
    # K = saved_hyperparameters["K"]
    # lambda_Sigma_x_inv = saved_hyperparameters["lambda_Sigma_x_inv"]

    trained_model = Popari(
        datasets=datasets,
        replicate_names=replicate_names,
        reloaded_hierarchy=reloaded_hierarchy,
        pretrained=True,
        # K=K,
        # metagene_mode=metagene_mode,
        # metagene_groups=metagene_groups,
        # spatial_affinity_groups=spatial_affinity_groups,
        # lambda_Sigma_x_inv=lambda_Sigma_x_inv,
        initial_context=context,
        torch_context=context,
        **new_kwargs,
    )

    return trained_model


def from_pretrained(pretrained_model: Popari, popari_context: dict = None, lambda_Sigma_bar: float = 1e-3):
    """Initialize Popari object from a SpiceMix pretrained model."""

    pretrained_datasets = pretrained_model.hierarchy[0].datasets
    datasets = [PopariDataset(dataset, dataset.name) for dataset in pretrained_datasets]
    replicate_names = [dataset.name for dataset in datasets]

    reloaded_hierarchy = None

    reloaded_hierarchy = {}
    for level in range(pretrained_model.hierarchical_levels):
        level_datasets = pretrained_model.hierarchy[level].datasets
        reloaded_hierarchy[level] = [
            PopariDataset(level_dataset, level_dataset.name) for level_dataset in level_datasets
        ]

    return load_pretrained(
        datasets,
        replicate_names,
        reloaded_hierarchy=reloaded_hierarchy,
        spatial_affinity_mode="differential lookup",
        context=popari_context,
        lambda_Sigma_bar=lambda_Sigma_bar,
    )



import argparse
import json

# has to be done at the end, after everything has been imported
import sys
from pathlib import Path

from popari.model import Popari, from_pretrained
from popari.train import Trainer, TrainParameters

from . import analysis as tl
from . import plotting as pl
from .__about__ import __version__

sys.modules.update({f"{__name__}.{m}": globals()[m] for m in ["tl", "pl"]})


def get_parser():
    """Get argument parser for Popari hyperparameter declaration."""

    parser = argparse.ArgumentParser(description="Run the Popari algorithm.")

    parser.add_argument(
        "--K",
        type=int,
        required=True,
        default=10,
        help="Number of metagenes to use for all replicates.",
    )
    parser.add_argument(
        "--num_iterations",
        type=int,
        required=True,
        default=200,
        help="Number of iterations to run Popari.",
    )
    parser.add_argument(
        "--nmf_preiterations",
        type=int,
        default=5,
        help="Number of NMF preiterations to use for initialization.",
    )
    parser.add_argument(
        "--dtype",
        type=str,
        required=True,
        default="float64",
        help="Datatype to use for PyTorch operations. Choose between ``float32`` or ``float64``",
    )
    parser.add_argument(
        "--torch_device",
        type=str,
        required=True,
        help="keyword args to use of PyTorch tensors during training.",
    )
    parser.add_argument(
        "--initial_device",
        type=str,
        required=True,
        help="keyword args to use during initialization of PyTorch tensors.",
    )
    parser.add_argument("--output_path", type=str, required=True, help="Path at which to save Popari output.")
    parser.add_argument("--dataset_path", type=str, required=True, help="Path to input dataset.")
    parser.add_argument(
        "--lambda_Sigma_x_inv",
        type=float,
        help="Hyperparameter to balance importance of spatial information.",
    )
    parser.add_argument("--pretrained", type=bool, help="if set, attempts to load model state from input files")
    parser.add_argument(
        "--initialization_method",
        type=str,
        help="algorithm to use for initializing metagenes and embeddings. Default ``svd``",
    )
    parser.add_argument(
        "--hierarchical_levels",
        type=int,
        help="Number of hierarchical binning levels to use in Popari run.",
    )
    parser.add_argument(
        "--binning_downsample_rate",
        type=float,
        help="Approximate rate at which spots are aggregated into bins.",
    )
    parser.add_argument(
        "--superresolution_lr",
        type=float,
        help="Initial learning rate for superresolution optimization.",
    )
    parser.add_argument(
        "--superresolution_epochs",
        type=int,
        default=10000,
        help="Number of epochs to do superresolution optimization.",
    )
    parser.add_argument(
        "--metagene_groups",
        type=json.loads,
        help="defines a grouping of replicates for the metagene optimization.",
    )
    parser.add_argument(
        "--spatial_affinity_groups",
        type=json.loads,
        help="defines a grouping of replicates for the spatial affinity optimization.",
    )
    parser.add_argument(
        "--betas",
        type=json.loads,
        help="weighting of each dataset during optimization. Defaults to equally weighting each dataset",
    )
    parser.add_argument(
        "--prior_x_modes",
        type=json.loads,
        help="family of prior distribution for embeddings of each dataset",
    )
    parser.add_argument("--M_constraint", type=str, help="constraint on columns of M. Default ``simplex``")
    parser.add_argument("--sigma_yx_inv_mode", type=str, help="form of sigma_yx_inv parameter. Default ``separate``")
    parser.add_argument(
        "--spatial_affinity_mode",
        type=str,
        help="modality of spatial affinity parameters. Default ``shared lookup``",
    )

    parser.add_argument(
        "--downsampling_method",
        type=str,
        help="Method to use for downsampling in hierarchical mode. Default ``partition``",
    )

    parser.add_argument("--metagene_mode", type=str, help="modality of metagene parameters. Default ``shared``")
    parser.add_argument(
        "--lambda_M",
        type=float,
        help="hyperparameter to constrain metagene deviation in differential case.",
    )
    parser.add_argument(
        "--lambda_Sigma_bar",
        type=float,
        help="hyperparameter to constrain spatial affinity deviation in differential case.",
    )
    parser.add_argument("--spatial_affinity_lr", type=float, help="learning rate for optimization of ``Sigma_x_inv``")
    parser.add_argument(
        "--spatial_affinity_tol",
        type=float,
        help="convergence tolerance during optimization of ``Sigma_x_inv``",
    )
    parser.add_argument(
        "--spatial_affinity_constraint",
        type=str,
        help="method to ensure that spatial affinities lie within an appropriate range",
    )
    parser.add_argument(
        "--spatial_affinity_centering",
        type=bool,
        help="if set, spatial affinities are zero-centered after every optimization step",
    )
    parser.add_argument(
        "--spatial_affinity_scaling",
        type=float,
        help="magnitude of spatial affinities during initial scaling. Default ``10``",
    )
    parser.add_argument(
        "--spatial_affinity_regularization_power",
        type=int,
        help="exponent controlling penalization of spatial affinity magnitudes. Default ``2``",
    )
    parser.add_argument(
        "--embedding_mini_iterations",
        type=int,
        help="number of mini-iterations to use during each iteration of embedding optimization. Default ``1000``",
    )
    parser.add_argument(
        "--embedding_acceleration_trick",
        type=bool,
        help="if set, use trick to accelerate convergence of embedding optimization. Default ``True``",
    )
    parser.add_argument(
        "--embedding_step_size_multiplier",
        type=float,
        help="controls relative step size during embedding optimization. Default ``1.0``",
    )
    parser.add_argument(
        "--use_inplace_ops",
        type=bool,
        help="if set, inplace PyTorch operations will be used to speed up computation",
    )
    parser.add_argument(
        "--random_state",
        type=int,
        help="seed for reproducibility of randomized computations. Default ``0``",
    )
    parser.add_argument(
        "--verbose",
        type=int,
        help="level of verbosity to use during optimization. Default ``0`` (no print statements)",
    )

    return parser


def main():
    parser = get_parser()

    args = parser.parse_args()
    filtered_args = {key: value for key, value in vars(args).items() if key is not None}

    num_iterations = filtered_args.pop("num_iterations")
    nmf_preiterations = filtered_args.pop("nmf_preiterations")

    torch_device = filtered_args.pop("torch_device")
    initial_device = filtered_args.pop("initial_device")
    dtype = filtered_args.pop("dtype")

    train_parameters = TrainParameters(
        nmf_iterations=nmf_iterations,
        nmf_preiterations=nmf_preiterations,
        savepath=output_path,
    )

    superresolution_epochs = filtered_args.pop("superresolution_epochs")

    dtype_object = torch.float32
    if dtype == "float64":
        dtype_object = torch.float64

    initial_context = {"device": initial_device, "dtype": dtype_object}
    filtered_args["initial_context"] = initial_context

    torch_context = {"device": torch_device, "dtype": dtype_object}
    filtered_args["torch_context"] = torch_context

    output_path = filtered_args.pop("output_path")
    output_path = Path(output_path)

    train_parameters = TrainParameters(
        iterations=num_iterations,
        nmf_iterations=nmf_preiterations,
        savepath=output_path,
    )

    model = Popari(**filtered_args)

    trainer = Trainer(
        parameters=parameters,
        model=model,
        verbose=filtered_args["verbose"],
    )

    trainer.train()
    trainer.superresolve(n_epochs=superresolution_epochs)
    trainer.save_results()



import datetime
import logging
import os
import pickle
import sys
import time
from collections import defaultdict
from typing import Optional, Sequence

import anndata as ad
import awkward as ak
import gseapy as gp
import matplotlib
import matplotlib.patches as patches
import networkx as nx
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
import torch
from anndata import AnnData
from kneed import KneeLocator
from matplotlib import pyplot as plt
from ortools.graph.python import min_cost_flow
from scipy.sparse import csr_array, sparray
from scipy.spatial.distance import pdist, squareform
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import accuracy_score, adjusted_rand_score, f1_score, silhouette_score
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import StandardScaler
from tqdm.auto import tqdm, trange
from umap import UMAP

from popari._popari_dataset import PopariDataset
from popari.sample_for_integral import integrate_of_exponential_over_simplex


def create_neighbor_groups(replicate_names, covariate_values, window_size=1):
    if window_size == None:
        return None

    unique_covariates = np.unique(covariate_values)
    covariate_to_replicate_name = defaultdict(list)
    for covariate_value, replicate_name in zip(covariate_values, replicate_names):
        covariate_to_replicate_name[covariate_value].append(replicate_name)

    sort_indices = np.argsort(unique_covariates)
    sorted_covariates = unique_covariates[sort_indices]

    groups = {}
    for index in range(window_size, len(sorted_covariates) - window_size):
        group_covariates = sorted_covariates[index - window_size : index + window_size + 1]
        group_replicates = sum(
            [covariate_to_replicate_name[group_covariate] for group_covariate in group_covariates],
            [],
        )
        group_name = f"{sorted_covariates[index]}"

        groups[group_name] = list(group_replicates)

    return groups


def concatenate(datasets: Sequence[PopariDataset], join: str = "inner", batch_key: str = "batch"):
    """Merge datasets in a way that is compatible with Popari.

    Args:
        datasets: list of PopariDataset.

    """
    dataset_names = [dataset.name for dataset in datasets]
    merged_dataset = ad.concat(
        datasets,
        label=batch_key,
        join=join,
        keys=dataset_names,
        merge="unique",
        uns_merge="unique",
        pairwise=True,
    )

    return merged_dataset


def unconcatenate(merged_dataset: ad.AnnData, batch_key: str = "batch"):
    """Unmerge concatenated."""

    indices = merged_dataset.obs.groupby(batch_key, observed=False).indices.values()
    datasets = [merged_dataset[index].copy() for index in indices]

    replicate_names = [dataset.obs[batch_key].unique()[0] for dataset in datasets]
    unmerged_datasets = [PopariDataset(dataset, name) for dataset, name in zip(datasets, replicate_names)]

    return unmerged_datasets


def calc_modularity(adjacency_matrix, label, resolution=1):
    adjacency_matrix = adjacency_matrix.tocoo()
    n = adjacency_matrix.shape[0]
    adjacency_matrix_sum = adjacency_matrix.data.sum()
    score = (
        adjacency_matrix.data[label[adjacency_matrix.row] == label[adjacency_matrix.col]].sum() / adjacency_matrix_sum
    )

    idx = np.argsort(label)
    label = label[idx]
    k = np.array(adjacency_matrix.sum(0)).ravel() / adjacency_matrix_sum
    k = k[idx]
    idx = np.concatenate([[0], np.nonzero(label[:-1] != label[1:])[0] + 1, [len(label)]])
    score -= sum(k[i:j].sum() ** 2 for i, j in zip(idx[:-1], idx[1:])) * resolution
    return score


def evaluate_embedding(obj, embedding="X", do_plot=True, do_sil=True):
    if embedding == "X":
        Xs = [X.cpu().numpy() for X in obj.Xs]
    elif embedding in obj.phenotype_predictors:
        Xs = [obj.phenotype_predictors["cell type encoded"][0](X).cpu().numpy() for X in obj.Xs]
    else:
        raise NotImplementedError

    # x = np.concatenate(Xs, axis=0)

    for x, replicate, dataset in zip(Xs, obj.repli_list, obj.datasets):
        x = StandardScaler().fit_transform(x)
        for n_clust in [6, 7, 8, 9]:
            y = AgglomerativeClustering(
                n_clusters=n_clust,
                linkage="ward",
            ).fit_predict(x)
            y = pd.Categorical(y, categories=np.unique(y))
            dataset.obs["spicemixplus_label"] = y

            print(
                f"replicate {replicate}"
                f"hierarchical w/ K={n_clust}."
                f"ARI = {adjusted_rand_score(dataset.obs['cell_type'].values, y):.2f}",
                sep=" ",
            )
        # y = clustering_louvain_nclust(
        #     x.copy(), 8,
        #     kwargs_neighbors=dict(n_neighbors=10),
        #     kwargs_clustering=dict(),
        #     resolution_boundaries=(.1, 1.),
        # )

    # obj.meta['label SpiceMixPlus'] = y
    # print('ari = {:.2f}'.format(adjusted_rand_score(*obj.meta[['cell type', 'label SpiceMixPlus']].values.T)))
    # for repli, df in obj.meta.groupby('repli'):
    #     print('ari {} = {:.8f}'.format(repli, adjusted_rand_score(*df[['cell type', 'label SpiceMixPlus']].values.T)))
    if do_plot:
        ncol = 4
        nrow = (obj.num_repli + ncol - 1) // ncol
        fig, axes = plt.subplots(nrow, ncol, figsize=(4 * ncol, 4 * nrow))
        for ax, replicate, dataset in zip(axes.flat, obj.repli_list, obj.datasets):
            sns.heatmap(
                dataset.obs.groupby(["cell_type", "spicemixplus_label"]).size().unstack().fillna(0).astype(int),
                ax=ax,
                annot=True,
                fmt="d",
            )
        plt.show()
        plt.close()

    if do_sil:
        for x, replicate, dataset in zip(Xs, obj.repli_list, obj.datasets):
            x = StandardScaler().fit_transform(x)
            x = UMAP(
                random_state=obj.random_state,
                n_neighbors=10,
            ).fit_transform(x)
            print("sil", silhouette_score(x, dataset.obs["cell_type"], random_state=obj.random_state))
            if do_plot:
                keys = ["cell_type", "replicate", "spicemixplus_label"]
                ncol = len(keys)
                nrow = 1 + obj.num_repli
                fig, axes = plt.subplots(nrow, ncol, figsize=(5 * ncol, 5 * nrow))

                def plot(axes, idx):
                    for ax, key in zip(axes, keys):
                        sns.scatterplot(ax=ax, data=dataset.obs.iloc[idx], x=x[idx, 0], y=x[idx, 1], hue=key, s=5)

                plot(axes[0], slice(None))
                for ax_row, (repli, df) in zip(axes[1:], dataset.obs.reset_index().groupby("replicate")):
                    plot(ax_row, df.index)
                plt.show()
                plt.close()


def evaluate_prediction_wrapper(obj, *, key_truth="cell type encoded", display_fn=print, **kwargs):
    X_all = torch.cat([X for X in obj.Xs], axis=0)
    obj.meta["label SpiceMixPlus predictor"] = obj.phenotype_predictors[key_truth][0](X_all).argmax(1).cpu().numpy()
    display_fn(evaluate_prediction(obj.meta, key_pred="label SpiceMixPlus predictor", key_truth=key_truth, **kwargs))


def evaluate_prediction(df_meta, *, key_pred="label SpiceMixPlus", key_truth="cell type", key_repli="repli"):
    df_score = {}
    for repli, df in df_meta.groupby(key_repli):
        t = tuple(df[[key_truth, key_pred]].values.T)
        r = {
            "acc": accuracy_score(*t),
            "f1 micro": f1_score(*t, average="micro"),
            "f1 macro": f1_score(*t, average="macro"),
            "ari": adjusted_rand_score(*t),
        }
        df_score[repli] = r
    df_score = pd.DataFrame(df_score).T
    return df_score


class NesterovGD:
    """Optimizer that implements Nesterov's Accelerated Gradient Descent.

    See below for implementation details:
    # https://blogs.princeton.edu/imabandit/2013/04/01/acceleratedgradientdescent/

    Attributes:
        parameters: object to optimize with Nesterov
        step_size: size of gradient update

    """

    def __init__(self, parameters: torch.Tensor, step_size: float):
        """Initialize Nesterov optimization.

        Args:
            parameters: object to optimize with Nesterov
            step_size: size of gradient update

        """
        self.parameters = parameters
        self.step_size = step_size
        # self.y = x.clone()
        self.y = torch.zeros_like(parameters)
        # self.lam = 0
        self.k = 0

    def set_parameters(self, parameters):
        """Reset parameters.

        Args:
            parameters: object to optimize with Nesterov

        """
        self.parameters = parameters

    def step(self, grad: torch.Tensor):
        """Update parameters according to state and step size.

        Args:
            grad: naive gradient for updating parameters

        """

        # method 1
        # lam_new = (1 + np.sqrt(1 + 4 * self.lam ** 2)) / 2
        # gamma = (1 - self.lam) / lam_new
        # method 2
        self.k += 1
        gamma = -(self.k - 1) / (self.k + 2)
        # method 3 - GD
        # gamma = 0
        # y_new = self.x.sub(grad, alpha=self.step_size)
        y_new = self.parameters - grad * self.step_size  # use addcmul
        self.y = (self.y * gamma).add(y_new, alpha=(1 - gamma))
        self.parameters = self.y
        # self.lam = lam_new
        self.y = y_new

        return self.parameters


def print_datetime():
    return datetime.datetime.now().strftime("[%Y/%m/%d %H:%M:%S]\t")


def get_datetime():
    return datetime.datetime.now().strftime("[%Y/%m/%d %H:%M:%S]\t")


def getRank(m, thr=0):
    rank = np.empty(m.shape, dtype=int)
    for r, a in zip(rank, m):
        r[np.argsort(a)] = np.arange(len(r))
        mask = a < thr
        r[mask] = np.mean(r[mask])
    return rank


@torch.no_grad()
def project_M(M, M_constraint):
    result = M.clone()
    if M_constraint == "simplex":
        result = project2simplex(result, dim=0, zero_threshold=1e-5)
    elif M_constraint == "unit sphere":
        result = M.div(torch.linalg.norm(result, ord=2, dim=0, keepdim=True))
    elif M_constraint == "nonneg unit sphere":
        result = M.clip(1e-10).div(torch.linalg.norm(result, ord=2, dim=0, keepdim=True))
    else:
        raise NotImplementedError
    return result


def project_M_(M, M_constraint):
    result = M.clone()
    if M_constraint == "simplex":
        result = project2simplex_(result, dim=0, zero_threshold=1e-5)
    elif M_constraint == "unit sphere":
        result = M.div(torch.linalg.norm(result, ord=2, dim=0, keepdim=True))
    elif M_constraint == "nonneg unit sphere":
        result = M.clip(1e-10).div(torch.linalg.norm(result, ord=2, dim=0, keepdim=True))
    else:
        raise NotImplementedError
    return result


def project2simplex(y, dim: int = 0, zero_threshold: float = 1e-10) -> torch.Tensor:
    """Projects a matrix such that the columns (or rows) lie on the unit
    simplex.

    See https://math.stackexchange.com/questions/2402504/orthogonal-projection-onto-the-unit-simplex
    for a reference.

    The goal is to find a scalar mu such that || (y-mu)_+ ||_1 = 1

    Currently uses Newton's method to optimize || y - mu ||^2

    TODO: try implementing it this way instead: https://arxiv.org/pdf/1101.6081.pdf

    Args:
        y: list of vectors to be projected to unit simplex
        dim: dimension along which to project
        zero_threshold: threshold to treat as zero for numerical stability purposes

    """

    num_components = y.shape[dim]

    mu = (y.sum(dim=dim, keepdim=True) - 1) / num_components
    previous_derivative = derivative = None
    for _ in range(num_components):
        difference = y - mu
        derivative = -(difference > zero_threshold).sum(dim=dim, keepdim=True).to(y.dtype)
        assert -derivative.min() > 0, difference.clip(min=0).sum(dim).min()
        if previous_derivative is not None and (derivative == previous_derivative).all():
            break
        objective_value = torch.clip(difference, min=zero_threshold).sum(dim=dim, keepdim=True) - 1
        newton_update = objective_value / derivative
        mu -= newton_update

        previous_derivative = derivative
    assert (derivative == previous_derivative).all()

    assert not torch.isnan(y).any(), y

    y = (y - mu).clip(min=zero_threshold)
    assert not torch.isnan(y).any(), (mu, derivative)

    assert y.sum(dim=dim).sub_(1).abs_().max() < 1e-3, y.sum(dim=dim).sub_(1).abs_().max()

    return y


def project2simplex_(y, dim: int = 0, zero_threshold: float = 1e-10) -> torch.Tensor:
    """(In-place) Projects a matrix such that the columns (or rows) lie on the
    unit simplex.

    See https://math.stackexchange.com/questions/2402504/orthogonal-projection-onto-the-unit-simplex
    for a reference.

    The goal is to find a scalar mu such that || (y-mu)_+ ||_1 = 1

    Currently uses Newton's method to optimize || y - mu ||^2

    TODO: try implementing it this way instead: https://arxiv.org/pdf/1101.6081.pdf

    Args:
        y: list of vectors to be projected to unit simplex
        dim: dimension along which to project
        zero_threshold: threshold to treat as zero for numerical stability purposes

    """
    y_copy = y.clone()
    num_components = y.shape[dim]

    y_copy.sub_(y_copy.sum(dim=dim, keepdim=True).sub_(1), alpha=1 / num_components)
    mu = y_copy.max(dim=dim, keepdim=True)[0].div_(2)
    derivative_prev, derivative = None, None
    for _ in range(num_components):
        difference = y_copy.sub(mu)
        objective_value = difference.clip_(min=zero_threshold).sum(dim, keepdim=True).sub_(1)
        derivative = difference.gt_(zero_threshold).sum(dim, keepdim=True)

        if derivative_prev is not None and (derivative == derivative_prev).all():
            break

        mu.addcdiv_(objective_value, derivative)
        derivative_prev = derivative

    y_copy.sub_(mu).clip_(min=zero_threshold)
    assert y_copy.sum(dim=dim).sub_(1).abs_().max() < 1e-4, y_copy.sum(dim=dim).sub_(1).abs_().max()
    return y_copy


class IndependentSet:
    """Iterator class that yields a list of batch_size independent nodes from a
    spatial graph.

    For each iteration, no pair of yielded nodes can be neighbors of each other according to the
    adjacency matrix.

    Attributes:
        N: number of nodes in graph
        adjacency_list: graph neighbor information stored in adjacency list format
        batch_size: number of nodes to draw independently every iteration

    """

    def __init__(self, adjacency_list, device, batch_size=50):
        self.N = len(adjacency_list)
        self.adjacency_list = adjacency_list
        self.batch_size = batch_size
        self.indices_remaining = None
        self.device = device

    def __iter__(self):
        """Return iterator over nodes in graph.

        Resets indices_remaining before returning the iterator.

        """
        self.indices_remaining = set(range(self.N))
        return self

    def __next__(self):
        """Returns the indices of batch_size nodes such that none of them are
        neighbors of each other.

        Makes sure selected nodes are not adjacent to each other, i.e., finds an
        independent set of `valid indices` in a greedy manner

        """
        if len(self.indices_remaining) == 0:
            raise StopIteration

        valid_indices = sample_graph_iid(self.adjacency_list, self.indices_remaining, self.batch_size)
        self.indices_remaining -= set(valid_indices)

        return torch.tensor(valid_indices, device=self.device, dtype=torch.long)


def sample_graph_iid(adjacency_list, indices_remaining, sample_size):
    valid_indices = []
    excluded_indices = set()
    effective_batch_size = min(sample_size, len(indices_remaining))
    candidate_indices = np.random.choice(
        list(indices_remaining),
        size=effective_batch_size,
        replace=False,
    )
    for index in candidate_indices:
        if index not in excluded_indices:
            valid_indices.append(index)
            excluded_indices |= set(adjacency_list[index])

    return valid_indices


def convert_numpy_to_pytorch_sparse_coo(numpy_coo, context):
    indices = np.array(numpy_coo.nonzero())
    values = numpy_coo.data[numpy_coo.data.nonzero()]

    i = torch.LongTensor(indices)
    v = torch.FloatTensor(values)
    size = numpy_coo.shape

    torch_coo = torch.sparse_coo_tensor(i, v, size=size, **context)

    return torch_coo


def compute_neighborhood_enrichment(features: np.ndarray, adjacency_matrix: csr_array):
    r"""Compute the normalized enrichment of features in direct neighbors on a
    graph.

    Args:
        features: attributes on the nodes of the graph on which to compute enrichment.
        adjacency_matrix: sparse graph representation.

    """

    adjacency_matrix = (adjacency_matrix + adjacency_matrix.T).astype(bool).astype(adjacency_matrix.dtype)
    edges_per_node = np.squeeze(np.asarray(adjacency_matrix.sum(axis=0)))
    connected_mask = edges_per_node > 0

    features = features[connected_mask]
    adjacency_matrix = adjacency_matrix[connected_mask][:, connected_mask]

    total_counts = features.sum(axis=0)[:, np.newaxis]
    assert np.all(total_counts > 0)

    normalized_enrichment = ((1 / total_counts) * features.T) @ (
        1 / edges_per_node[connected_mask][:, np.newaxis] * adjacency_matrix.toarray() @ features
    )

    return np.asarray(normalized_enrichment)


def bin_expression(
    spot_expression: np.ndarray,
    spot_coordinates: np.ndarray,
    bin_coordinates: np.ndarray,
    num_jobs: int,
):
    """Bin spot expressions into filtered coordinates.

    Args:
        spot_expression: expression of original spots
        spot_coordinates: coordinates of original spots
        bin_coordinates: coordinates of downsampled bin spots

    Returns:
        A tuple of (binned expression for metaspots, assignment matrix of bins to spots)

    """

    num_spots, num_genes = spot_expression.shape
    num_bins, _ = bin_coordinates.shape

    bin_expression = np.zeros((num_bins, num_genes))
    bin_assignments = csr_array((num_bins, num_spots), dtype=np.int32)

    neigh = NearestNeighbors(n_neighbors=1, n_jobs=num_jobs)
    neigh.fit(bin_coordinates)
    indices = neigh.kneighbors(spot_coordinates, return_distance=False)

    bin_to_spots = defaultdict(list)
    for i in range(num_spots):
        bin_to_spots[indices[i].item()].append(i)

    for i in range(num_bins):
        bin_spots = bin_to_spots[i]
        bin_assignments[i, bin_spots] = 1
        bin_expression[i] = np.sum(spot_expression[bin_spots], axis=0)

    return bin_expression, bin_assignments


def expression_score(dataset, expression_key: str = "X", threshold: float = 99.0):
    """Expression score for spatial domain detection."""

    expression = dataset.obsm[expression_key]
    expression_threshold = np.percentile(expression, threshold, axis=0)
    mask = expression > expression_threshold

    total_entities = mask.sum(axis=0)
    total_expression = (expression_threshold * mask).sum(axis=0)
    expression_score = total_expression / total_entities

    dataset.uns["expression_score"] = {
        "threshold": threshold,
        "scores": expression_score,
    }

    return expression_score


def normalize_expression_by_threshold(dataset, thresholded_key: str = "elbowed_X", threshold: float = 99.0):
    """Replacement for Z-score threshold."""

    thresholded_expression = dataset.obsm[thresholded_key]
    expression_threshold = np.percentile(thresholded_expression, threshold, axis=0)
    mask = thresholded_expression > expression_threshold

    total_entities = mask.sum(axis=0)
    total_expression = (expression_threshold * mask).sum(axis=0)

    normalized_thresholded_expression = thresholded_expression / total_expression

    dataset.obsm["normalized_thresholded_expression"] = normalized_thresholded_expression

    return normalized_thresholded_expression


def smooth_metagene_expression(
    dataset,
    processed_key: str = "normalized_thresholded_expression",
    adjacency_list_key: str = "adjacency_list",
):
    """"""
    adjacency_list = dataset.obsm[adjacency_list_key]

    processed_expression = dataset.obsm[processed_key]
    smoothed_expression = np.zeros_like(processed_expression)
    for entity in np.arange(len(dataset)):
        adjacencies = adjacency_list[entity]
        neighbor_expressions = processed_expression[adjacencies]
        average_expression = (processed_expression[entity] + neighbor_expressions.sum(axis=0)) / (
            len(neighbor_expressions) + 1
        )
        smoothed_expression[entity] = average_expression

    dataset.obsm["smoothed_expression"] = smoothed_expression

    return smoothed_expression


def spatially_smooth_feature(labels, adjacency_list, max_smoothing_rounds=10, smoothing_threshold=0.5):
    """"""
    num_entities = len(labels)

    smoothed_labels = labels.copy()
    for _ in range(max_smoothing_rounds):
        new_labels = smoothed_labels.copy()
        for entity in np.arange(num_entities):
            current_cluster = labels[entity]

            adjacencies = adjacency_list[entity]
            neighbor_labels = labels[adjacencies]
            num_neighbors = len(neighbor_labels)
            if num_neighbors == 0:
                new_labels[entity] = current_cluster
                continue

            values, counts = np.unique(neighbor_labels, return_counts=True)

            max_index = np.argmax(counts)
            max_cluster = values[max_index]

            ratio = (counts[max_index] + (max_cluster == current_cluster)) / (num_neighbors + 1)
            if ratio >= smoothing_threshold:
                new_labels[entity] = max_cluster
            else:
                new_labels[entity] = current_cluster

        if np.all(smoothed_labels == new_labels):
            break

        smoothed_labels = new_labels

    return new_labels


def smooth_labels(
    dataset,
    label_key: str = "leiden",
    output_key: str = "smoothed_leiden",
    smoothing_threshold: float = 0.5,
    max_smoothing_rounds: int = 10,
    adjacency_list_key: str = "adjacency_list",
):
    """"""
    adjacency_list = dataset.obsm[adjacency_list_key]

    labels = dataset.obs[label_key]
    dataset.obs[output_key] = spatially_smooth_feature(
        labels,
        adjacency_list,
        max_smoothing_rounds,
        smoothing_threshold,
    )

    return dataset.obs[output_key]


def run_gsea(
    gene_list: Sequence[str],
    name: str,
    background: Sequence[str],
    output_name=None,
    mode: str = "dotplot",
    **enrichr_kwargs,
):
    """GSEApy analysis.

    Args:
        gene_list: list of gene names to check for enrichment
        background: list of background genes to use to for comparison
        name: title for analysis plot
        output_name: path where plot figure will be saved
        enrichr_kwargs: keyword arguments for the call to `gp.enrichr`
        mode: what type of plot to produce. Default: `"dotplot"`

    """

    organism = enrichr_kwargs.pop("organism", "mouse")
    gene_sets = enrichr_kwargs.pop("gene_sets", ["GO_Biological_Process_2023"])

    enrichment_result = gp.enrichr(
        gene_list=gene_list,
        gene_sets=gene_sets,
        organism=organism,
        background=background,
        outdir=None,
        **enrichr_kwargs,
    )

    enrichment_result.results.sort_values(by="Adjusted P-value")

    if mode == "dotplot":
        ax = gp.dotplot(
            enrichment_result.results,
            column="Adjusted P-value",
            x="Gene_set",  # set x axis, so you could do a multi-sample/library comparsion
            size=2,
            top_term=5,
            figsize=(3, 5),
            title=f"{name} GSEA Enrichment",
            xticklabels_rot=45,  # rotate xtick labels
            show_ring=True,  # set to False to revmove outer ring
            ofname=output_name,
            marker="o",
        )
    elif mode == "barplot":
        ax = gp.barplot(
            enrichment_result.results,
            column="Adjusted P-value",
            group="Gene_set",  # set group, so you could do a multi-sample/library comparsion
            size=10,
            top_term=5,
            figsize=(3, 5),
            color=["red", "green", "blue"],
            title=f"{name} GSEA Enrichment",
            ofname=output_name,
        )

    return enrichment_result, ax


def get_metagene_signature(
    metagene,
    gene_names,
    sensitivity: float = 1.0,
    type: str = "upregulated",
    show_plot: bool = False,
):
    """Use knee-detection algorithm to get top genes for metagene."""

    num_genes = len(metagene)

    sort_indices = np.argsort(metagene)
    curve = "convex" if type == "upregulated" else "concave"

    kneedle = KneeLocator(range(num_genes), metagene[sort_indices], S=sensitivity, curve=curve, direction="increasing")

    signature_range = slice(kneedle.knee, None) if type == "upregulated" else slice(None, kneedle.knee)
    signature_genes = gene_names[sort_indices[signature_range]]

    if show_plot:
        kneedle.plot_knee()

    return list(signature_genes)


def convert_adjacency_matrix_to_awkward_array(adjacency_matrix: sparray):
    """Convert COO adjacency matrix to ragged Awkward Array."""

    num_cells, _ = adjacency_matrix.shape
    adjacency_list = [[] for _ in range(num_cells)]
    for x, y in zip(*adjacency_matrix.nonzero()):
        adjacency_list[x].append(y)

    adjacency_list = ak.Array(adjacency_list)

    return adjacency_list



import json
import pickle as pkl
import random
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Sequence, Tuple, Union

import anndata as ad
import numpy as np
import scanpy as sc
import scipy
import squidpy as sq
from anndata import AnnData
from matplotlib import pyplot as plt
from numpy.typing import NDArray
from scipy.spatial.distance import cdist
from scipy.stats import gamma, truncnorm

from popari._canvas import DomainCanvas, MetageneCanvas
from popari.util import convert_adjacency_matrix_to_awkward_array


def sample_gaussian(sigma: NDArray, means: NDArray, N: int = 1, random_state=0) -> NDArray:
    """Sample multivariate Gaussian given a mean vector and covariance matrix.

    Args:
        sigma: covariance matrix
        means: mean vector
        N: number of multivariate samples to take

    Returns:
        (K, N) sample matrix

    """

    rng = np.random.default_rng(random_state)
    K = len(sigma)
    assert sigma.shape[0] == sigma.shape[1]
    assert len(means) == K

    # Box-Muller Method
    L = np.linalg.cholesky(sigma)
    n_z = K + (K % 2)
    x = np.zeros((n_z, N))
    num_samples = 0
    while True:
        n_valid = 0
        while True:
            z = 2 * rng.random(2) - 1
            if z[0] ** 2 + z[1] ** 2 <= 1:
                r = np.linalg.norm(z)
                x[n_valid, num_samples] = z[0] * np.sqrt(-2 * np.log(r**2) / r**2)
                x[n_valid + 1, num_samples] = z[1] * np.sqrt(-2 * np.log(r**2) / r**2)
                n_valid += 2
            if n_valid == n_z:
                num_samples += 1
                break
        if num_samples == N:
            break

    # if K is odd, there will be one extra sample, so throw it away
    x = x[0:K, :]
    x = np.dot(L, x) + np.expand_dims(means, -1)

    return np.squeeze(x)


def sample_2D_points(num_points, minimum_distance: float, width: float = 1.0, height: float = 1.0, random_state=0):
    """Generate 2D samples that are at least minimum_distance apart from each
    other."""
    # TODO: Implement Poisson disc sampling for a vectorized operation
    rng = np.random.default_rng(random_state)

    points = np.zeros((num_points, 2))
    points[0] = rng.random(2) * np.array([width, height])
    for index in range(1, num_points):
        while True:
            point = rng.random((1, 2)) * np.array([width, height])
            distances = cdist(points[:index], point)
            if np.min(distances) > minimum_distance:
                points[index] = point
                break

    return points


def synthesize_metagenes_nsf(
    num_genes,
    num_spatial_metagenes: int,
    n_nonspatial_metagenes: int,
    spatial_metagene_parameter: float,
    nonspatial_metagene_parameter: float,
    original_metagenes: Optional[NDArray] = None,
    normalize: bool = False,
    random_state: int = 0,
    nonspatial_nonzero_prob: Optional[float] = None,
):
    """Test nsf metagene synthesis."""
    # TODO: delete the below lines for generalizability
    rng = np.random.default_rng(random_state)

    num_metagenes = num_spatial_metagenes + n_nonspatial_metagenes
    metagenes = np.zeros((num_metagenes, num_genes))

    # last_index = None
    metagene_indices = rng.choice(num_spatial_metagenes, size=num_genes, replace=True)
    metagenes[metagene_indices, np.arange(num_genes)] = spatial_metagene_parameter

    if n_nonspatial_metagenes > 0:
        if nonspatial_nonzero_prob is not None:
            for metagene_index in range(n_nonspatial_metagenes):
                gene_indices = rng.binomial(n=1, p=nonspatial_nonzero_prob, size=num_genes).astype(bool)
                metagenes[num_spatial_metagenes + metagene_index, gene_indices] = nonspatial_metagene_parameter
        else:
            metagene_indices = rng.choice(n_nonspatial_metagenes, size=num_genes, replace=True)
            metagenes[metagene_indices + num_spatial_metagenes, np.arange(num_genes)] = nonspatial_metagene_parameter

    return metagenes


def sample_normalized_embeddings(Z: NDArray, sigma_x: NDArray, rng: Union[int, np.random.Generator]):
    """Sample embeddings from truncated Gaussian given mean vectors, and project
    to simplex.

    Args:
        Z: mean values for each emebdding dimension
        sigma_x: variance vector for all embedding dimensions

    """

    rng = np.random.default_rng(rng)

    X = np.zeros_like(Z)
    num_cells, num_metagenes = Z.shape
    # TODO: vectorize
    for cell in range(num_cells):
        for metagene in range(num_metagenes):
            X[cell, metagene] = (
                sigma_x[metagene] * truncnorm.rvs(-Z[cell, metagene] / sigma_x[metagene], 100, random_state=rng)
                + Z[cell, metagene]
            )

    X = X * (Z > 0)
    X = X / np.sum(X, axis=1, keepdims=True)

    return X


def synthesize_cell_embeddings_nsf(
    cell_type_labels,
    num_cells,
    cell_type_definitions,
    num_spatial_metagenes,
    n_nonspatial_metagenes=3,
    signal_sigma_x=0.1,
    background_sigma_x=0.2,
    nonspatial_nonzero_prob=0.1,
    sigma_x_scale=1.0,
    random_state=None,
):
    """Generate synthetic cell embeddings."""

    rng = np.random.default_rng(random_state)
    num_metagenes = num_spatial_metagenes + n_nonspatial_metagenes

    _, cell_type_encoded_labels = np.unique(cell_type_labels, return_inverse=True)
    cell_type_assignments = np.zeros((num_cells), dtype="int")
    Z = np.zeros((num_cells, num_metagenes))

    for cell_index, (cell_type, cell_type_encoded) in enumerate(zip(cell_type_labels, cell_type_encoded_labels)):
        cell_type_assignments[cell_index] = cell_type_encoded
        Z[cell_index, :num_spatial_metagenes] = cell_type_definitions[cell_type]

    # Extrinsic factors
    Z[:, num_spatial_metagenes:num_metagenes] = rng.binomial(
        1,
        nonspatial_nonzero_prob,
        size=(num_cells, n_nonspatial_metagenes),
    )
    Z += 1e-6

    return Z, cell_type_assignments


@dataclass
class SimulationParameters:
    """Container for simulation parameters.

    Args:
        num_cells: number of cells to simulate
        num_genes: number of total genes to simulate
        annotation_mode: whether the Canvas is annotating `domain` or `cell_type`
        num_real_metagenes: number of real metagenes
        num_noise_metagenes: number of noise metagenes
        real_metagene_parameter: shape parameter for Gamma distribution from which real metagene
            weights are sampled
        noise_metagene_parameter: shape parameter for Gamma distribution from which noise metagene
            weights are sampled
        spatial_distributions: proportions of cell types in each domain of the simulation
        cell_type_definitions: definitions of simulated cell types by metagene proportion
        metagene_variation_probabilities: variation of metagene weight definitions between metagenes
        domain_key: key in `.obs` where the domain identity of each cell is stored
        width: width of the canvas
        height: height of the canvas
        minimum_distance: minimum distance between simulated datapoints
        grid_size: number of rows/columns for grid-based simulation. Alternative to `num_cells` parameter.
        sig_y_scale: standard deviation of additive Gaussian noise used during sampling of gene expression
        sig_x_scale: standard deviation of additive Gaussian noise used during sampling of latent states
        lambda_s: shape parameter used for sampling cell sizes

    """

    num_cells: Optional[int] = None
    num_genes: int = 100
    annotation_mode: str = "domain"
    num_real_metagenes: int = 10
    num_noise_metagenes: int = 3
    real_metagene_parameter: float = 4.0
    noise_metagene_parameter: float = 4.0
    spatial_distributions: dict = None
    cell_type_definitions: dict = None
    metagene_variation_probabilities: Sequence = None
    domain_key: str = "domain"
    width: float = 1.0
    height: float = 1.0
    minimum_distance: float = None
    grid_size: int = None
    sig_y_scale: float = 3.0
    sig_x_scale: float = 3.0
    lambda_s: float = 1.0


def get_grid_coordinates(width: float, height: float, grid_size: int):
    """Generate grid coordinates for simulation.

    Assumes exactly grid_size rows and grid_size columns, therefore yielding (grid_size)^2 rows.

    Args:
        width: width of grid, in relative units
        height: height of grid, in relative units
        grid_size: number of rows/columns in the grid.

    """

    x = np.linspace(0, width, grid_size)
    y = np.linspace(0, height, grid_size)
    xv, yv = np.meshgrid(x, y)

    coordinates = np.vstack([xv.flatten() * width, yv.flatten() * height]).T

    return coordinates


class SyntheticDataset(AnnData):
    """Simulated spatial transcriptomics dataset.

    Uses AnnData as a base class, with additional methods for simulation.

    """

    def __init__(
        self,
        replicate_name: Union[int, str],
        parameters: SimulationParameters,
        random_state: Union[int, np.random.Generator] = None,
        verbose: int = 0,
    ):
        """Generate random coordinates (as well as expression values) for a
        single ST FOV."""

        self.params = parameters
        self.verbose = verbose

        try:
            self.params.num_cells = self.params.num_cells or self.params.grid_size**2
        except TypeError as e:
            raise ValueError(
                "At least one of `num_cells` or `grid_size` must be defined in the input `SimulationParameters` dataclass object.",
            )

        invalid_definition_lengths = [
            (len(cell_type_definition) != self.params.num_real_metagenes)
            for cell_type_definition in self.params.cell_type_definitions.values()
        ]
        if (len(self.params.metagene_variation_probabilities) != self.params.num_real_metagenes) or np.any(
            invalid_definition_lengths,
        ):
            raise ValueError(
                "The dimensions of simulation parameters must be aligned. Please "
                "check that the `metagene_variation_probabilities`, `cell_type_definitions`"
                " and `num_real_metagenes` values are all compatible.",
            )

        self.rng = np.random.default_rng(random_state)

        dummy_expression = np.zeros((self.params.num_cells, self.params.num_genes))

        super().__init__(X=dummy_expression)

        self.name = f"{replicate_name}"

        self.uns["domain_names"] = list(self.params.spatial_distributions.keys())

        self.uns["simulation_parameters"] = vars(parameters)

        if self.params.annotation_mode == "domain":
            self.uns["simulation_parameters"]["cell_type_names"] = list(self.params.cell_type_definitions.keys())
            self.uns["cell_type_definitions"] = {
                self.name: self.params.cell_type_definitions,
            }

        if self.params.grid_size:
            self.obsm["spatial"] = get_grid_coordinates(self.params.width, self.params.height, self.params.grid_size)
        else:
            if not self.params.minimum_distance:
                minimum_distance = 0.75 / np.sqrt(self.params.num_cells)
            tau = minimum_distance * 2.2
            self.obsm["spatial"] = sample_2D_points(
                self.params.num_cells,
                minimum_distance,
                width=self.params.width,
                height=self.params.height,
                random_state=self.rng,
            )

        canvas_constructor = DomainCanvas if self.params.annotation_mode == "domain" else MetageneCanvas
        self.domain_canvas = canvas_constructor(
            self.obsm["spatial"],
            self.uns["domain_names"],
            canvas_width=600,
            density=1,
        )

    def synthesize_metagenes(
        self,
        original_metagenes: Optional[NDArray] = None,
        replicate_variability: Optional[float] = None,
        normalize: bool = True,
    ) -> NDArray:
        """Synthesize related metagenes according to the
        `metagene_variation_probabilities` vector.

        Creates `num_real_metagenes` synthetic metagenes using a random Gamma distribution with
        shape parameter `real_metagene_parameter`. For each metagene `i`, if
        `dropout_probabilities[i] != 0`, randomly permutes a `metagene_variation_probabilities[i]`
        fraction of metagene `i-1` to create metagene `i`; otherwise, creates a new random metagene.
        In addition, adds `n_noise_metagenes` parameterized by a Gamma distribution with shape
        parameter `noise_metagene_parameter`.

        Args:
            original_metagenes: base metagenes to resample. If `None`, metagenes are generated from
                scratch.
            replicate_variability: variation in metagenes definition between simulation replicates
            normalize: whether metagenes should be normalized such that each metagene sums to 1.
            random_state: random state to use for sampling of metagene weights

        Return:
            An array of ground truth metagene weights.

        """

        num_metagenes = self.params.num_real_metagenes + self.params.num_noise_metagenes
        metagenes = np.zeros((num_metagenes, self.params.num_genes))

        #     last_index = None
        for index in range(self.params.num_real_metagenes):
            variation_probability = self.params.metagene_variation_probabilities[index]

            if variation_probability == 0 and not replicate_variability:
                metagene = gamma.rvs(
                    self.params.real_metagene_parameter,
                    size=self.params.num_genes,
                    random_state=self.rng,
                )
                metagenes[index] = metagene
            #             last_index = index
            else:
                if variation_probability == 0:
                    metagene = original_metagenes[index].copy()
                    variation_probability = replicate_variability

                mask = np.full(self.params.num_genes, False)
                masked_genes = self.rng.choice(
                    self.params.num_genes,
                    size=int(variation_probability * self.params.num_genes),
                    replace=False,
                )
                mask[masked_genes] = True
                perturbed_metagene = metagene.copy()

                perturbations = gamma.rvs(self.params.real_metagene_parameter, size=np.sum(mask), random_state=self.rng)
                if original_metagenes is not None:
                    # Use dirichlet distribution
                    perturbations *= np.sum(metagene[mask]) / np.sum(perturbations)
                perturbed_metagene[mask] = perturbations

                metagenes[index] = perturbed_metagene

        #         print(f"Difference between last_index and current index: {((metagenes[index] - metagenes[last_index]) == 0).sum() / num_genes}")

        for index in range(self.params.num_real_metagenes, num_metagenes):
            metagenes[index] = gamma.rvs(
                self.params.noise_metagene_parameter,
                size=self.params.num_genes,
                random_state=self.rng,
            )

        metagenes = metagenes

        if normalize:
            metagenes = metagenes / np.sum(metagenes, axis=1, keepdims=True)

        return metagenes

    def synthesize_cell_embeddings(self, signal_sigma_x=0.1, background_sigma_x=0.03):
        """Generate synthetic cell embeddings."""

        domain_labels = self.obs[self.params.domain_key].to_numpy()

        num_metagenes = self.params.num_real_metagenes + self.params.num_noise_metagenes

        cell_type_assignments = np.zeros((self.params.num_cells), dtype="int")
        Z = np.zeros((self.params.num_cells, num_metagenes))

        cell_types = self.params.cell_type_definitions.keys()

        for domain_index, (domain_name, distribution) in enumerate(self.params.spatial_distributions.items()):
            domain_cells = domain_labels == domain_name
            domain_cell_types, proportions = zip(*distribution.items())

            (cell_indices,) = np.nonzero(domain_cells)
            self.rng.shuffle(cell_indices)

            partition_indices = (np.cumsum(proportions) * len(cell_indices)).astype(int)
            partitions = np.split(cell_indices, partition_indices[:-1])

            cell_type_to_partition = dict(zip(domain_cell_types, partitions))

            for cell_type_index, cell_type in enumerate(cell_types):
                if cell_type not in domain_cell_types:
                    continue

                partition = cell_type_to_partition[cell_type]
                if len(partition) == 0:
                    continue

                cell_type_assignments[partition] = cell_type_index
                Z[partition, : self.params.num_real_metagenes] = self.params.cell_type_definitions[cell_type]

        # Extrinsic factors
        Z[:, self.params.num_real_metagenes : num_metagenes] = 0.05

        sigma_x = np.concatenate(
            [
                np.full(self.params.num_real_metagenes, signal_sigma_x),
                np.full(self.params.num_noise_metagenes, background_sigma_x),
            ],
        )
        sigma_x = sigma_x * self.params.sig_x_scale

        X = sample_normalized_embeddings(Z, sigma_x, rng=self.rng)

        return X, cell_type_assignments

    def simulate_expression(self, predefined_metagenes=None, metagene_magnitudes=None, **simulation_parameters):
        """Simulate expression using parameters."""
        if self.verbose:
            print(f"Simulating {self.params.annotation_mode}-annotated expression...")

        if self.params.annotation_mode == "domain":
            self.simulate_metagene_based_expression(
                predefined_metagenes=predefined_metagenes,
                metagene_magnitudes=metagene_magnitudes,
            )
        elif self.params.annotation_mode == "metagene":
            return simulate_nsf_expression(predefined_metagenes=predefined_metagenes, **simulation_parameters)

    def simulate_metagene_based_expression(
        self,
        metagene_magnitudes=None,
        predefined_metagenes=None,
    ):
        """Simulate metagenes and embeddings following metagene-based SpiceMix
        model."""

        # Get num_genes x num_genes covariance matrix
        num_metagenes = self.params.num_real_metagenes + self.params.num_noise_metagenes
        if isinstance(self.params.sig_y_scale, float):
            self.variance_y = (
                self.params.sig_y_scale * np.identity(self.params.num_genes) / self.params.num_genes
            ) ** 2
        elif isinstance(self.params.sig_y_scale, dict):
            self.variance_y = {
                cell_type: cell_specific_sig_y_scale / self.params.num_genes
                for cell_type, cell_specific_sig_y_scale in self.params.sig_y_scale.items()
            }
            random_key = next(iter(self.variance_y))
            if isinstance(self.variance_y[random_key], float):
                self.variance_y = {
                    cell_type: cell_specific_sig_y_scale * np.identity(self.params.num_genes)
                    for cell_type, cell_specific_sig_y_scale in self.variance_y.items()
                }

            self.variance_y = {
                cell_type: cell_specific_sig_y_scale**2
                for cell_type, cell_specific_sig_y_scale in self.variance_y.items()
            }

        if self.verbose > 1:
            print(f"Gene covariance set: {self.variance_y is not None}")

        magnitudes = None
        if predefined_metagenes is None:
            metagenes = self.synthesize_metagenes()

            self.magnitudes = np.sum(metagenes, axis=1)
            metagenes = metagenes / self.magnitudes[:, np.newaxis]
            metagenes = metagenes.T
        else:
            metagenes = predefined_metagenes
            self.magnitudes = metagene_magnitudes

        self.uns["ground_truth_M"] = {self.name: metagenes}

        X_i, C_i = self.synthesize_cell_embeddings()

        self.S = gamma.rvs(
            num_metagenes / self.params.lambda_s,
            scale=self.params.lambda_s,
            size=self.params.num_cells,
            random_state=self.rng,
        )
        self.obsm["ground_truth_X"] = X_i * self.S[:, np.newaxis]
        cell_type_encoded = C_i.astype(int)
        cell_type = [self.uns["simulation_parameters"]["cell_type_names"][index] for index in cell_type_encoded]
        self.obs["cell_type"] = cell_type
        self.obs["cell_type_encoded"] = cell_type_encoded

        self.sample_noisy_expression()

    def simulate_nsf_expression(
        self,
        num_spatial_metagenes: int,
        num_nonspatial_metagenes: int,
        spatial_metagene_parameter: float,
        nonspatial_metagene_parameter: float,
        lambda_s: float,
        background_expression: float = 0.2,
        predefined_metagenes=None,
        metagene_magnitudes=None,
        rate: float = 10.0,
        nonspatial_nonzero_prob: Optional[float] = None,
    ):
        """Simulate metagenes and embeddings following metagene-based NSF model.

        Args:
            num_spatial_metagenes: Number of spatial metagenes to simulate
            num_nonspatial_metagenes: Number of non-spatial (intrinsic) metagenes to simulate

        """

        num_metagenes = num_spatial_metagenes + num_nonspatial_metagenes
        magnitudes = None
        if predefined_metagenes is None:
            metagenes = synthesize_metagenes_nsf(
                self.num_genes,
                num_spatial_metagenes,
                num_nonspatial_metagenes,
                spatial_metagene_parameter,
                nonspatial_metagene_parameter,
                random_state=self.rng,
                normalize=True,
                nonspatial_nonzero_prob=nonspatial_nonzero_prob,
            )

            self.magnitudes = np.sum(metagenes, axis=1)
            metagenes = metagenes / self.magnitudes[:, np.newaxis]
            metagenes = metagenes.T
        else:
            metagenes = predefined_metagenes
            self.magnitudes = metagene_magnitudes

        self.uns["ground_truth_M"] = {self.name: metagenes}

        X_i, C_i = synthesize_cell_embeddings_nsf(
            self.obs["cell_type"].to_numpy(),
            self.num_cells,
            cell_type_definitions=self.uns["cell_type_definitions"][self.name],
            num_spatial_metagenes=num_spatial_metagenes,
            n_nonspatial_metagenes=num_nonspatial_metagenes,
            random_state=self.rng,
        )

        # self.S = gamma.rvs(num_metagenes, scale=lambda_s, size=self.num_cells)
        self.obsm["ground_truth_X"] = X_i * self.magnitudes

        # self.obsm["ground_truth_X"] = X_i

        cell_type_encoded = C_i.astype(int)
        cell_type = [self.uns["cell_type_names"][index] for index in cell_type_encoded]
        self.obs["cell_type"] = cell_type
        self.obs["cell_type_encoded"] = cell_type_encoded

        self.sample_noisy_expression(background_expression=background_expression, rate=rate)

    def sample_noisy_expression(self, background_expression: float = 0.2, rate=10.0):
        """Samples gene expression from Negative Binomial distribution according
        to SyntheticDataset attributes.

        Uses ``self.variance_y``.

        TODO: make it so that mean expression is stored in .obsm, not in .X (so that this method does
        not mutate mean expression).

        """

        self.X = np.matmul(self.obsm["ground_truth_X"], self.uns["ground_truth_M"][self.name].T)

        if self.params.annotation_mode == "domain":
            for cell, cell_type in zip(range(self.params.num_cells), self.obs["cell_type_encoded"].to_numpy()):
                if isinstance(self.variance_y, dict):
                    cell_type_variance_y = self.variance_y[int(cell_type)]
                else:
                    cell_type_variance_y = self.variance_y

                # Ensure that gene expression is positive
                self.X[cell] = np.abs(sample_gaussian(cell_type_variance_y, self.X[cell], random_state=self.rng))

        elif self.annotation_mode == "metagene":
            self.X += background_expression
            self.X = self.rng.negative_binomial(rate, rate / (self.X + rate))

            self.raw = self
            sc.pp.log1p(self)

    def annotate_domain(self, points=None):
        return self.domain_canvas.annotate_domain(points=points)

    def assign_domain_labels(self):
        if self.params.annotation_mode == "metagene":
            self.domain_canvas.convert_metagenes_to_cell_types()

        domain_kd_tree = self.domain_canvas.generate_domain_kd_tree()
        self.obs[self.params.domain_key] = domain_kd_tree.query(self.obsm["spatial"])
        if self.params.annotation_mode == "metagene":
            self.uns["cell_type_names"] = list(set(self.obs[self.params.domain_key]))
            cell_type_definitions = {
                label: np.zeros(len(self.domain_canvas.domain_names)) for label in self.uns["cell_type_names"]
            }

            self.uns["cell_type_definitions"] = {
                self.name: cell_type_definitions,
            }
            for cell_type in self.uns["cell_type_names"]:
                definition = json.loads(cell_type)
                if None not in definition:
                    self.uns["cell_type_definitions"][self.name][cell_type][definition] = 1

        self.uns["domain_landmarks"] = dict(self.domain_canvas.domains)


class MultiReplicateSyntheticDataset:
    """Synthetic multireplicate dataset to model biological variation and batch
    effects in spatial transcriptomics data."""

    def __init__(
        self,
        replicate_parameters: dict[SimulationParameters],
        dataset_constructor: SyntheticDataset,
        random_state=0,
        verbose=0,
    ):
        self.verbose = verbose
        self.datasets = {}
        self.replicate_parameters = replicate_parameters

        # random.seed(random_state)
        self.rng = np.random.default_rng(random_state)

        for replicate_name in self.replicate_parameters:
            synthetic_dataset = dataset_constructor(
                replicate_name=replicate_name,
                parameters=replicate_parameters[replicate_name],
                random_state=self.rng,
                verbose=self.verbose,
            )
            self.datasets[replicate_name] = synthetic_dataset

    def annotate_replicate_domain(self, replicate_name, points=None):
        print(f"Annotating replicate {replicate_name}")

        return self.datasets[replicate_name].annotate_domain(points=points)

    def __iter__(self):
        yield from self.datasets.values()

    def assign_domain_labels(self):
        for replicate_dataset in self:
            replicate_dataset.assign_domain_labels()

    def simulate_expression(self):
        metagenes = None
        metagene_magnitudes = None
        for replicate_dataset in self:
            if self.verbose:
                print(f"Simulating expression for {replicate_dataset.name}.")

            replicate_dataset.simulate_expression(
                predefined_metagenes=metagenes,
                metagene_magnitudes=metagene_magnitudes,
            )

            metagenes = replicate_dataset.uns["ground_truth_M"][replicate_dataset.name]
            metagene_magnitudes = replicate_dataset.magnitudes

    def calculate_neighbors(self, **neighbors_kwargs):
        coord_type = neighbors_kwargs.pop("coord_type", "generic")
        delaunay = neighbors_kwargs.pop("delaunay", True)
        for dataset in self:
            sq.gr.spatial_neighbors(
                dataset,
                coord_type=coord_type,
                delaunay=delaunay,
                radius=[0, 0.1],
                **neighbors_kwargs,
            )
            dataset.obsp["adjacency_matrix"] = dataset.obsp["spatial_connectivities"]
            dataset.obsm["adjacency_list"] = convert_adjacency_matrix_to_awkward_array(dataset.obsp["adjacency_matrix"])



import argparse
import json
import traceback
from pathlib import Path

import mlflow
import torch
from matplotlib import pyplot as plt

from popari import get_parser, pl, tl
from popari.model import Popari
from popari.train import MLFlowTrainer, MLFlowTrainParameters


def main():
    parser = get_parser()

    parser.add_argument(
        "--spatial_preiterations",
        type=int,
        default=0,
        help="Number of SpiceMix preiterations to use for initialization.",
    )

    parser.add_argument(
        "--save_figs",
        type=bool,
        default=True,
        help="Whether or not to save intermediate figs.",
    )

    args = parser.parse_args()
    filtered_args = {key: value for key, value in vars(args).items() if value is not None}

    print(filtered_args)

    num_iterations = filtered_args.pop("num_iterations")
    nmf_preiterations = filtered_args.pop("nmf_preiterations")
    spatial_preiterations = filtered_args.pop("spatial_preiterations")
    save_figs = filtered_args.pop("save_figs")

    torch_device = filtered_args.pop("torch_device")
    initial_device = filtered_args.pop("initial_device")
    dtype = filtered_args.pop("dtype")
    filtered_args["verbose"] -= 1

    superresolution_epochs = filtered_args.pop("superresolution_epochs", 0)

    dtype_object = torch.float32
    if dtype == "float64":
        dtype_object = torch.float64

    initial_context = {"device": initial_device, "dtype": dtype_object}
    filtered_args["initial_context"] = initial_context

    torch_context = {"device": torch_device, "dtype": dtype_object}
    filtered_args["torch_context"] = torch_context

    output_path = filtered_args.pop("output_path")
    output_path = Path(output_path)

    checkpoint_iterations = 100

    train_parameters = MLFlowTrainParameters(
        iterations=num_iterations,
        nmf_iterations=nmf_preiterations,
        spatial_preiterations=spatial_preiterations,
        save_figs=save_figs,
        savepath=output_path,
    )

    model = Popari(**filtered_args)

    mlflow_trainer = MLFlowTrainer(
        parameters=train_parameters,
        model=model,
        verbose=filtered_args["verbose"] + 1,
    )

    with mlflow_trainer:
        if "metagene_groups" in filtered_args:
            mlflow.set_tag("disjoint_metagenes", (filtered_args["metagene_groups"] == "disjoint"))
        if "spatial_affinity_groups" in filtered_args:
            mlflow.set_tag("disjoint_spatial_affinities", (filtered_args["spatial_affinity_groups"] == "disjoint"))

        trackable_hyperparameters = (
            "K",
            "lambda_Sigma_bar",
            "lambda_Sigma_x_inv",
            "hierarchical_levels",
            "downsampling_method",
            "binning_downsample_rate",
            "spatial_affinity_mode",
            "random_state",
            "dataset_path",
        )

        mlflow.log_params(
            {
                **{
                    hyperparameter: filtered_args[hyperparameter]
                    for hyperparameter in trackable_hyperparameters
                    if hyperparameter in filtered_args
                },
                "nmf_preiterations": nmf_preiterations,
                "spatial_preiterations": spatial_preiterations,
                "num_iterations": num_iterations,
            },
        )

        try:
            mlflow_trainer.train()
            mlflow_trainer.superresolve(n_epochs=superresolution_epochs)
            mlflow_trainer.save_results()

        except Exception as e:
            with open(Path(f"./output_{torch_device}.txt"), "a") as f:
                tb = traceback.format_exc()
                f.write(f"\n{tb}")
        finally:
            if Path(f"./output_{torch_device}.txt").is_file():
                mlflow.log_artifact(f"output_{torch_device}.txt")


if __name__ == "__main__":
    main()



from functools import partial
from typing import Optional, Sequence

import numpy as np
import scanpy as sc
from matplotlib import cm
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import ListedColormap

from popari._dataset_utils import (
    _evaluate_classification_task,
    _multigroup_heatmap,
    _multireplicate_heatmap,
    _plot_all_embeddings,
    _plot_cell_type_to_metagene,
    _plot_cell_type_to_metagene_difference,
    _plot_clusters_to_categories,
    _plot_confusion_matrix,
    _plot_in_situ,
    _plot_metagene_embedding,
    _plot_metagene_signature_enrichment,
    _plot_umap,
    for_model,
    setup_squarish_axes,
)
from popari.model import Popari

in_situ = for_model(_plot_in_situ, return_outputs=True)
metagene_embedding = for_model(_plot_metagene_embedding, return_outputs=True)
confusion_matrix = for_model(_plot_confusion_matrix, return_outputs=True)
umap = for_model(_plot_umap, return_outputs=True)
multireplicate_heatmap = for_model(_multireplicate_heatmap, return_outputs=True)
clusters_to_categories = for_model(_plot_clusters_to_categories, return_outputs=True)
metagene_signature_enrichment = for_model(_plot_metagene_signature_enrichment, return_outputs=True)


def multigroup_heatmap(
    trained_model: Popari,
    title_font_size: Optional[int] = None,
    group_type: str = "metagene",
    axes: Optional[Sequence[Axes]] = None,
    key: Optional[str] = None,
    level=0,
    **heatmap_kwargs,
):
    r"""Plot 2D heatmap data across all datasets.

    Wrapper function to enable plotting of continuous 2D data across multiple replicates. Only
    one of ``obsm``, ``obsp`` or ``uns`` should be used.

    Args:
        trained_model: the trained Popari model.
        axes: A predefined set of matplotlib axes to plot on.
        obsm: the key in the ``.obsm`` dataframe to plot.
        obsp: the key in the ``.obsp`` dataframe to plot.
        uns: the key in the ``.uns`` dataframe to plot. Unstructured data must be 2D in shape.
        **heatmap_kwargs: arguments to pass to the `ax.imshow` call for each dataset

    """
    datasets = trained_model.hierarchy[level].datasets

    groups = trained_model.metagene_groups if group_type == "metagene" else trained_model.spatial_affinity_groups
    _multigroup_heatmap(datasets, title_font_size=title_font_size, groups=groups, axes=axes, key=key, **heatmap_kwargs)


def spatial_affinities(
    trained_model: Popari,
    title_font_size: Optional[int] = None,
    spatial_affinity_key: Optional[str] = "Sigma_x_inv",
    axes: Optional[Sequence[Axes]] = None,
    level=0,
    **heatmap_kwargs,
):
    r"""Plot Sigma_x_inv across all datasets.

    Wrapper function to enable plotting of continuous 2D data across multiple replicates. Only
    one of ``obsm``, ``obsp`` or ``uns`` should be used.

    Args:
        trained_model: the trained Popari model.
        axes: A predefined set of matplotlib axes to plot on.
        obsm: the key in the ``.obsm`` dataframe to plot.
        obsp: the key in the ``.obsp`` dataframe to plot.
        uns: the key in the ``.uns`` dataframe to plot. Unstructured data must be 2D in shape.
        **heatmap_kwargs: arguments to pass to the `ax.imshow` call for each dataset

    """
    datasets = trained_model.hierarchy[level].datasets

    # Override following kwargs with
    cmap = heatmap_kwargs.pop("cmap") if "cmap" in heatmap_kwargs else "bwr"
    nested = heatmap_kwargs.pop("nested") if "nested" in heatmap_kwargs else True
    max_value = round(
        np.max(np.abs(np.array([dataset.uns[spatial_affinity_key][dataset.name] for dataset in datasets]))),
    )
    vmin = -max_value
    vmax = max_value

    _multireplicate_heatmap(
        datasets,
        title_font_size=title_font_size,
        axes=axes,
        uns=spatial_affinity_key,
        nested=nested,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        **heatmap_kwargs,
    )


def all_embeddings(
    trained_model: Popari,
    embedding_key: str = "X",
    column_names: Optional[str] = None,
    level=0,
    **spatial_kwargs,
):
    r"""Plot all learned metagenes in-situ across all replicates.

    Each replicate's metagenes are contained in a separate plot.

    Args:
        trained_model: the trained Popari model.
        embedding_key: the key in the ``.obsm`` dataframe for the cell/spot embeddings.
        column_names: a list of the suffixes for each latent feature. If ``None``, it is assumed
            that these suffixes are just the indices of the latent features.

    """

    datasets = trained_model.hierarchy[level].datasets

    first_dataset = datasets[0]
    _, K = first_dataset.obsm[f"{embedding_key}"].shape

    if column_names == None:
        column_names = [f"{embedding_key}_{index}" for index in range(K)]

    _plot_all_embeddings(datasets, embedding_key=embedding_key, column_names=column_names, **spatial_kwargs)


def cell_type_to_metagene(trained_model: Popari, cell_type_de_genes: dict, level=0, **correspondence_kwargs):
    r"""Plot distribution of gene ranks of marker genes within each metagene.

    Args:
        trained_model: the trained Popari model.
        cell_type_de_genes: dictionary mapping each cell type to a list of marker genes.

    Returns:
        mapping from each cell type to the median rank of its marker genes in each metagene

    """

    datasets = trained_model.hierarchy[level].datasets

    first_dataset = datasets[0]

    fig, medians = _plot_cell_type_to_metagene(first_dataset, cell_type_de_genes, **correspondence_kwargs)

    return medians, fig


def cell_type_to_metagene_difference(
    trained_model: Popari,
    cell_type_de_genes: dict,
    first_metagene: int,
    second_metagene: int,
    level=0,
    **correspondence_kwargs,
):
    r"""Plot distribution of gene ranks of marker genes within each metagene.

    Args:
        trained_model: the trained Popari model.
        cell_type_de_genes: dictionary mapping each cell type to a list of marker genes.

    Returns:
        mapping from each cell type to the median rank of its marker genes in each metagene

    """

    datasets = trained_model.hierarchy[level].datasets

    first_dataset = datasets[0]

    fig, medians = _plot_cell_type_to_metagene_difference(
        first_dataset,
        cell_type_de_genes,
        first_metagene,
        second_metagene,
        **correspondence_kwargs,
    )


def affinity_magnitude_vs_difference(
    trained_model,
    level=0,
    spatial_affinity_key: str = "Sigma_x_inv",
    spatial_affinity_bar_key: str = "spatial_affinity_bar",
    joint=False,
    figsize=(10, 10),
    n_best: int = 5,
):
    """Plot all pairwise affinities, in terms of absolute and relative
    magnitude.

    Args:
        trained_model: the trained Popari model.

    """
    datasets = trained_model.hierarchy[level].datasets
    group_suffix = f"_level_{level}" if level > 0 else ""

    fig, axes = setup_squarish_axes(len(datasets), figsize=figsize)

    all_top_pairs = []

    for index in range(len(datasets), axes.size):
        axes.flat[index].axis("off")

    for ax, (index, dataset) in zip(axes.flat, enumerate(datasets)):
        dataset.uns["delta_Sigma"] = {
            dataset.name: dataset.uns[spatial_affinity_key][dataset.name]
            - dataset.uns[spatial_affinity_bar_key][f"_default{group_suffix}"],
        }

        Sigma_x_inv = dataset.uns[spatial_affinity_key][dataset.name]
        delta_Sigma = dataset.uns["delta_Sigma"][dataset.name]
        pairs = {}
        for i in range(trained_model.K):
            for j in range(i + 1):
                pairs[(i, j)] = (delta_Sigma[(i, j)], Sigma_x_inv[(i, j)])

        indices, flat_pairs = zip(*pairs.items())

        magnitudes = np.linalg.norm(flat_pairs, axis=1)

        sorted_index = np.argsort(magnitudes)
        best_index = sorted_index[-n_best:][::-1]

        x, y = np.array(flat_pairs).T

        ax.scatter(x, y, s=1, color="#D3D3D3")

        num_top_points = abs(n_best)
        colors = sc.pl.palettes.godsnot_102[:num_top_points]

        best_indices = np.array(indices)[best_index]
        for (i, j), color in zip(best_indices, colors):
            x, y = pairs[(i, j)]
            ax.scatter(x, y, s=20, color=color, label=f"m{i} × m{j}")

        ax.set_title("Pairwise affinity scatter")
        ax.set_xlabel("Difference from average affinity")
        ax.set_ylabel("Pairwise affinity")
        ax.legend(loc="upper left", bbox_to_anchor=(1, 1))

        all_top_pairs.append(best_indices)

    return fig, all_top_pairs


def normalized_affinity_trends(
    trained_model,
    timepoint_values: Sequence[float],
    time_unit="Days",
    normalize: bool = False,
    spatial_affinity_key: str = "Sigma_x_inv",
    n_best: int = 5,
    highlight_metric: str = "pearson",
    figsize: tuple = None,
    margin_size: float = 0.25,
    level=0,
):
    """Plot trends for every pair of affinities; highlight top trends.

    Args:
        trained_model: the trained Popari model.
        timepoint_values: x-values against which to plot trends
        time_unit: unit in which time is measured (used for x-axis label)

    """

    datasets = trained_model.hierarchy[level].datasets

    first_dataset = datasets[0]
    spatial_trends = first_dataset.uns["spatial_trends"]

    all_affinities = np.array([dataset.uns[spatial_affinity_key][dataset.name] for dataset in datasets])

    if normalize:
        for index in range(len(datasets), axes.size):
            axes.flat[index].axis("off")
            prenormalization_affinity_std = np.std(all_affinities, axis=0, keepdims=True)
            prenormalization_timepoint_std = np.std(timepoint_values)
            all_affinities /= prenormalization_affinity_std
            timepoint_values /= prenormalization_timepoint_std

    timepoint_min = np.min(timepoint_values)
    timepoint_ptp = np.ptp(timepoint_values)
    timepoint_std = np.std(timepoint_values)

    affinity_min = np.min(all_affinities)
    affinity_ptp = np.ptp(all_affinities)
    affinity_std = np.std(all_affinities, axis=0, keepdims=True)

    if figsize is None:
        figsize = (10, 5)

    fig, ax = plt.subplots(dpi=300, figsize=figsize)

    ax.set_ylim([affinity_min - margin_size, affinity_min + affinity_ptp + margin_size])
    ax.set_xlim([timepoint_min - margin_size, timepoint_min + timepoint_ptp + margin_size])
    ax.set_xticks(timepoint_values)

    number_of_lines = abs(n_best)
    colors = cm.get_cmap("rainbow", number_of_lines)

    for i in range(trained_model.K):
        for j in range(i + 1):
            affinity_values = all_affinities[:, i, j]
            if [i, j] not in spatial_trends["top_pairs"]:
                #             if True:
                line = ax.plot(timepoint_values, affinity_values, color="#D3D3D3", linestyle="--", linewidth=0.5)

    for index, (i, j) in enumerate(spatial_trends["top_pairs"]):
        affinity_values = all_affinities[:, i, j]
        if highlight_metric == "pearson":
            r = spatial_trends["pearson_correlations"][(i, j)]
            slope = spatial_trends["slopes"][(i, j)]
            slope_display = f", slope={slope:.2f}" if not normalize else ""
            label = f"m{i} × m{j}, r={r:.2}{slope_display}"
        elif highlight_metric == "variance":
            variance = spatial_trends["variances"][(i, j)]
            label = f"m{i} × m{j}, σ={variance:.2f}"

        color = colors(index)
        ax.plot(timepoint_values, affinity_values, color=color, linestyle="-", linewidth=3, label=label, zorder=2)

    ax.set_title("Pairwise affinity trends")
    ax.set_xlabel(f"{time_unit}")
    ax.set_ylabel("Pairwise affinity")
    ax.legend(loc="upper left", bbox_to_anchor=(1, 0))

    return fig



from dataclasses import dataclass, field
from pathlib import Path

from matplotlib import pyplot as plt

try:
    import mlflow
except ImportError:
    pass

from tqdm.auto import trange

from popari import analysis as tl
from popari import plotting as pl
from popari.model import Popari


@dataclass
class TrainParameters:
    nmf_iterations: int
    iterations: int
    savepath: Path
    synchronization_frequency: int = field(default=10, kw_only=True)


@dataclass
class MLFlowTrainParameters:
    nmf_iterations: int
    spatial_preiterations: int
    iterations: int
    savepath: Path
    synchronization_frequency: int = field(default=50, kw_only=True)
    checkpoint_iterations: int = field(default=50, kw_only=True)
    save_figs: bool = field(default=False, kw_only=True)


class Trainer:
    def __init__(self, parameters: TrainParameters, model: Popari, verbose: int = 0):
        self.model = model
        self.parameters = parameters
        self.verbose = verbose
        self.nmf_iterations = 0
        self.iterations = 0

    def train(self):
        nmf_progress_bar = trange(self.parameters.nmf_iterations, leave=True, disable=not self.verbose)
        for _ in nmf_progress_bar:
            if self.verbose > 0:
                description = f"-------------- NMF Iteration {self.nmf_iterations} --------------"
                nmf_progress_bar.set_description(description)

            synchronize = not (self.nmf_iterations % self.parameters.synchronization_frequency)
            self.model.estimate_parameters(update_spatial_affinities=False, synchronize=synchronize)
            self.model.estimate_weights(use_neighbors=False, synchronize=synchronize)

            self.nmf_iterations += 1

        progress_bar = trange(self.parameters.iterations, leave=True, disable=not self.verbose)
        for _ in progress_bar:
            if self.verbose > 0:
                description = f"------------------ Iteration {self.iterations} ------------------"
                progress_bar.set_description(description)

            synchronize = not (self.iterations % self.parameters.synchronization_frequency)

            self.model.estimate_parameters(synchronize=synchronize)
            self.model.estimate_weights(synchronize=synchronize)

            self.iterations += 1

    def save_results(self, savepath=None, **kwargs):
        if savepath is None:
            savepath = self.parameters.savepath

        self.model.save_results(savepath, **kwargs)

    def superresolve(self, **kwargs):
        new_lr = kwargs.pop("new_lr", self.model.superresolution_lr)
        target_level = kwargs.pop("target_level", None)
        self.model.set_superresolution_lr(new_lr, target_level)
        self.model.superresolve(**kwargs)

    # TODO: decide whether to separate out saving of model training hyperparameters and
    # Popari parameters saving completely. This will obvious imply huge changes with how
    # MLflow should work. Perhaps even justifies switching entirely to WandB (a good excuse)


class MLFlowTrainer(Trainer):
    def __init__(self, parameters: MLFlowTrainParameters, model: Popari, verbose: int = 0):
        super().__init__(parameters, model, verbose)
        self.spatial_preiterations = 0
        self.torch_device = self.model.context["device"]

        self.is_hierarchical = self.model.hierarchical_levels > 1

        if not self.is_hierarchical:
            path_without_extension = self.parameters.savepath.parent / self.parameters.savepath.stem
            self.parameters.savepath = f"{path_without_extension}.h5ad"

    def train(self):
        nll = self.model.nll()
        mlflow.log_metric("nll", nll, step=-1)
        nmf_progress_bar = trange(self.parameters.nmf_iterations, leave=True, disable=not self.verbose)
        for _ in nmf_progress_bar:
            if self.verbose > 0:
                description = f"-------------- NMF Iteration {self.nmf_iterations} --------------"
                nmf_progress_bar.set_description(description)

            synchronize = not (self.nmf_iterations % self.parameters.synchronization_frequency)
            self.model.estimate_parameters(
                update_spatial_affinities=False,
                differentiate_metagenes=False,
                synchronize=synchronize,
            )
            self.model.estimate_weights(use_neighbors=False, synchronize=synchronize)

            self.nmf_iterations += 1

        nll = self.model.nll()
        mlflow.log_metric("nll", nll, step=self.nmf_iterations)

        self.model.parameter_optimizer.reinitialize_spatial_affinities()
        self.model.synchronize_datasets()

        nll_spatial = self.model.nll(use_spatial=True)
        mlflow.log_metric("nll_spatial", nll_spatial, step=-1)

        if self.parameters.save_figs:
            for hierarchical_level in range(self.model.hierarchical_levels):
                self.save_popari_figs(level=hierarchical_level, save_spatial_figs=False)

        self.save_results(self.parameters.savepath)

        spatial_preprogress_bar = trange(self.parameters.spatial_preiterations, leave=True, disable=not self.verbose)
        for _ in spatial_preprogress_bar:
            if self.verbose > 0:
                description = f"-------------- Spatial Preiteration {self.spatial_preiterations} --------------"
                spatial_preprogress_bar.set_description(description)

            synchronize = not (self.spatial_preiterations % self.parameters.synchronization_frequency)
            self.model.estimate_parameters(differentiate_spatial_affinities=False, synchronize=synchronize)
            self.model.estimate_weights(synchronize=synchronize)

            if self.spatial_preiterations % self.parameters.checkpoint_iterations == 0:
                nll_spatial = self.model.nll(use_spatial=True)
                mlflow.log_metric("nll_spatial_preiteration", nll_spatial, step=self.spatial_preiterations)

                if self.parameters.save_figs:
                    for hierarchical_level in range(self.model.hierarchical_levels):
                        self.save_popari_figs(level=hierarchical_level, save_spatial_figs=True)

                if Path(f"./output_{self.torch_device}.txt").is_file():
                    mlflow.log_artifact(f"output_{self.torch_device}.txt")

                self.save_results(self.parameters.savepath)

            self.spatial_preiterations += 1

        progress_bar = trange(self.parameters.iterations, leave=True, disable=not self.verbose)
        for _ in progress_bar:
            if self.verbose > 0:
                description = f"------------------------- Iteration {self.iterations} -------------------------"
                progress_bar.set_description(description)

            synchronize = not (self.iterations % self.parameters.synchronization_frequency)

            self.model.estimate_parameters(synchronize=synchronize)
            self.model.estimate_weights(synchronize=synchronize)

            if self.iterations % self.parameters.checkpoint_iterations == 0:
                checkpoint_path = f"{self.torch_device}_checkpoint_{self.iterations}_iterations"
                if not self.is_hierarchical:
                    checkpoint_path = f"{checkpoint_path}.h5ad"

                nll_spatial = self.model.nll(use_spatial=True)
                mlflow.log_metric("nll_spatial", nll_spatial, step=self.iterations)

                if self.parameters.save_figs:
                    for hierarchical_level in range(self.model.hierarchical_levels):
                        self.save_popari_figs(level=hierarchical_level, save_spatial_figs=True)

                if Path(f"./output_{self.torch_device}.txt").is_file():
                    mlflow.log_artifact(f"output_{self.torch_device}.txt")

                self.save_results(checkpoint_path)

            self.iterations += 1

        self.save_results(self.parameters.savepath, ignore_raw_data=False)

        if self.nmf_iterations + self.iterations > 0 and self.parameters.save_figs:
            for hierarchical_level in range(self.model.hierarchical_levels):
                self.save_popari_figs(level=hierarchical_level, save_spatial_figs=True)

    def superresolve(self, **kwargs):
        super().superresolve(**kwargs)

        if self.nmf_iterations + self.iterations > 0 and self.parameters.save_figs:
            for hierarchical_level in range(self.model.hierarchical_levels):
                self.save_popari_figs(level=hierarchical_level, save_spatial_figs=True)

    def __enter__(self, log_system_metrics: bool = True):
        return mlflow.start_run(log_system_metrics=log_system_metrics)

    def __exit__(self, error_type, value, traceback):
        mlflow.end_run()

    def save_results(self, savepath=None, **kwargs):
        if savepath is None:
            savepath = self.parameters.savepath

        super().save_results(savepath=savepath, **kwargs)
        mlflow.log_artifact(savepath)

    def save_popari_figs(self, level: int = 0, save_spatial_figs: bool = False):
        """Save Popari figures."""

        if not self.is_hierarchical:
            suffix = ".png"
        else:
            suffix = f"_level_{level}.png"

        tl.preprocess_embeddings(self.model, level=level)
        tl.leiden(self.model, level=level, joint=True)

        leiden_fig, axes = setup_squarish_axes(len(self.model.datasets))
        pl.in_situ(self.model, level=level, color="leiden", edges_width=0, axes=axes)

        leiden_fig.savefig(f"leiden{suffix}")
        mlflow.log_artifact(f"leiden{suffix}")

        if save_spatial_figs:
            pl.spatial_affinities(self.model, level=level)

            plt.savefig(f"Sigma_x_inv{suffix}")
            mlflow.log_artifact(f"Sigma_x_inv{suffix}")

            # pl.multireplicate_heatmap(
            #     self.model,
            #     level=level,
            #     uns="M",
            #     aspect=self.model.K / self.model.datasets[0].shape[1],
            #     cmap="hot",
            # )

            # plt.savefig(f"metagenes{suffix}")
            # mlflow.log_artifact(f"metagenes{suffix}")

        for metagene in range(self.model.K):
            pl.metagene_embedding(self.model, metagene, level=level)
            plt.savefig(f"metagene_{metagene}_in_situ{suffix}")
            mlflow.log_artifact(f"metagene_{metagene}_in_situ{suffix}")



import argparse
import itertools
import json
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import mlflow
import mlflow.projects
import mlflow.tracking
import numpy as np
import toml
from mlflow.tracking import MlflowClient

from popari.mlflow.util import generate_mlproject_file


def run():
    parser = argparse.ArgumentParser(description="Run Popari hyperparameter grid search.")
    parser.add_argument("--configuration_filepath", type=str, required=True, help="Path to configuration file.")
    parser.add_argument("--mlflow_output_path", type=str, default=".", help="Where to output MLflow results.")
    parser.add_argument("--debug", action="store_true", help="Whether to dump print statements to console or to file.")
    # parser.add_argument('--include_benchmarks', action='store_true', help="Whether to include NMF and SpiceMix benchmarks for each hyperparameter combination.")
    parser.add_argument(
        "--benchmarks",
        action="extend",
        nargs="+",
        default=[],
        type=str,
        help="Which benchmarks to include.",
    )

    parser.add_argument("--only_benchmarks", action="store_true", help="Whether to only run benchmarks.")

    args = parser.parse_args()

    with open(args.configuration_filepath) as f:
        configuration = toml.load(f)

    mlflow_output_path = Path(args.mlflow_output_path)
    generate_mlproject_file(configuration["runtime"]["project_name"], mlflow_output_path / "MLproject")

    hyperparameter_names = [
        "K",
        "lambda_Sigma_x_inv",
        "lambda_Sigma_bar",
        "random_state",
        "hierarchical_levels",
        "binning_downsample_rate",
        "nmf_preiterations",
        "num_iterations",
        "spatial_preiterations",
        "save_figs",
        "metagene_groups",
        "spatial_affinity_groups",
        "dtype",
        "downsampling_method",
    ]

    if "dataset_paths" in configuration["runtime"]:
        dataset_paths = configuration["runtime"]["dataset_paths"]
    elif "dataset_path" in configuration["runtime"]:
        dataset_paths = [configuration["runtime"]["dataset_path"]]

    configuration["hyperparameters"]["dataset_path"] = dataset_paths

    # Fix this up so that you can input a list of CUDA devices instead of a number of processes
    if "device_list" in configuration["runtime"]:
        device_list = configuration["runtime"]["device_list"]
        num_processes = len(device_list)
    elif "num_processes" in configuration["runtime"]:
        num_processes = configuration["runtime"]["num_processes"]
        device_list = [f"cuda:{device_number}" for device_number in range(num_processes)]

    tracking_client = MlflowClient()

    device_status = [False for _ in range(num_processes)]

    def generate_evaluate_function(parent_run, experiment_id, null_nll=0):
        """Generates function to evaluate Popari."""

        def evaluate(params):
            """Start parallel Popari run and track progress."""

            device_index = device_status.index(False)
            device = device_list[device_index]
            device_status[device_index] = True

            child_run = tracking_client.create_run(experiment_id, tags={"mlflow.parentRunId": parent_run.info.run_id})
            # with mlflow.start_run(nested=True) as child_run:
            p = mlflow.projects.run(
                run_id=child_run.info.run_id,
                uri=str(mlflow_output_path),
                entry_point="train_debug" if args.debug else "train",
                parameters={
                    **{
                        parameter_name: params[parameter_name]
                        for parameter_name in params
                        if params[parameter_name] is not None
                    },
                    "spatial_affinity_mode": (
                        "shared lookup" if params["lambda_Sigma_bar"] == 0 else "differential lookup"
                    ),
                    # "dataset_path": configuration['runtime']['dataset_path'],
                    "output_path": f"./device_{device}_result",
                    "torch_device": device,
                    "initial_device": device,
                    "spatial_affinity_mode": (
                        "shared lookup" if params["lambda_Sigma_bar"] == 0 else "differential lookup"
                    ),
                    "verbose": 2,
                },
                env_manager="local",
                experiment_id=experiment_id,
                synchronous=False,
            )
            succeeded = p.wait()
            device_status[device_index] = False

            tracking_client.set_terminated(child_run.info.run_id)

            if succeeded:
                training_run = tracking_client.get_run(p.run_id)
                metrics = training_run.data.metrics

                if "nll" in metrics:
                    nll = metrics["nll"]
                else:
                    raise RuntimeError(
                        "A run failed during initialization. This likely points "
                        "to an improperly formatted grid search configuration file.",
                    )
            else:
                tracking_client.set_terminated(p.run_id, "FAILED")
                nll = null_nll

            mlflow.log_metrics({"nll": nll})

            return p.run_id, nll

        return evaluate

    with mlflow.start_run() as parent_run:
        experiment_id = parent_run.info.experiment_id
        null_evaluate = generate_evaluate_function(parent_run, experiment_id)
        null_hyperparameters = {
            "K": configuration["hyperparameters"]["K"]["start"],
            "lambda_Sigma_x_inv": 0,
            "lambda_Sigma_bar": 0,
            "random_state": 0,
            "hierarchical_levels": 2,
            "binning_downsample_rate": 0.1,
            "nmf_preiterations": 0,
            "num_iterations": 0,
            "dtype": "torch32",
            "spatial_preiterations": 0,
            "save_figs": False,
            "metagene_groups": json.dumps(None),
            "spatial_affinity_groups": json.dumps(None),
            "downsampling_method": "partition",
            "dataset_path": dataset_paths[0],
        }
        _, null_nll = null_evaluate(null_hyperparameters)

        benchmarks = {
            "nmf_benchmark": {
                "nmf_preiterations": 200,
                "spatial_preiterations": 0,
                "num_iterations": 0,
                "lambda_Sigma_bar": 0,
                "lambda_Sigma_x_inv": 0,
            },
            "spicemix_benchmark": {
                "nmf_preiterations": 5,
                "spatial_preiterations": 200,
                "num_iterations": 0,
                "lambda_Sigma_bar": 0,
            },
            "disjoint_spicemix_benchmark": {
                "nmf_preiterations": 5,
                "spatial_preiterations": 200,
                "num_iterations": 0,
                "lambda_Sigma_bar": 0,
                "metagene_groups": json.dumps("disjoint"),
                "spatial_affinity_groups": json.dumps("disjoint"),
            },
        }

        benchmarks = {name: params for name, params in benchmarks.items() if name in args.benchmarks}

        hyperparameter_options_list = []
        for hyperparameter_name in hyperparameter_names:
            if hyperparameter_name not in configuration["hyperparameters"]:
                default_options = [null_hyperparameters[hyperparameter_name]]
                hyperparameter_options_list.append(default_options)
                continue

            search_space = configuration["hyperparameters"][hyperparameter_name]

            dtype = search_space["dtype"]

            # Categorical hyperparameters are specified via the "options" field
            if dtype == "categorical":
                search_space_options = search_space["options"]
                if hyperparameter_name.endswith("groups"):
                    search_space_options = [json.dumps(option) for option in search_space_options]

                hyperparameter_options = search_space_options
                hyperparameter_options_list.append(hyperparameter_options)
                continue

            # Handling numerical hyperparameters
            start = search_space["start"]
            end = search_space["end"]
            scale = search_space["scale"]
            gridpoints = search_space["gridpoints"]

            if scale == "log":
                gridspace = np.logspace
            elif scale == "linear":
                gridspace = np.linspace

            hyperparameter_options = gridspace(start, end, num=gridpoints)

            if dtype == "int":
                hyperparameter_options = np.rint(hyperparameter_options).astype(int)

            hyperparameter_options_list.append(hyperparameter_options)

        hyperparameter_names.append("dataset_path")
        hyperparameter_options_list.append(dataset_paths)

        
        options = list(
            dict(zip(hyperparameter_names, hyperparameter_choice))
            for hyperparameter_choice in itertools.product(*hyperparameter_options_list)
        )

        special_options = []
        for benchmark_name, benchmark in benchmarks.items():
            is_novel = [
                True if hyperparameter_name in benchmark else False for hyperparameter_name in hyperparameter_names
            ]

            special_options_list = [
                hyperparameter_options if not is_novel[index] else [benchmark[hyperparameter_name]]
                for index, (hyperparameter_name, hyperparameter_options) in enumerate(
                    zip(hyperparameter_names, hyperparameter_options_list),
                )
            ]
            special_options.extend(
                list(
                    dict(zip(hyperparameter_names, hyperparameter_choice))
                    for hyperparameter_choice in itertools.product(*special_options_list)
                ),
            )

        if args.only_benchmarks:
            options = special_options
        else:
            options.extend(special_options)

        print(f"Number of grid search candidates: {len(options)}")

        evaluate = generate_evaluate_function(parent_run, experiment_id, null_nll)
        with ThreadPoolExecutor(max_workers=num_processes) as executor:
            _ = executor.map(evaluate, options)

        # find the best run, log its metrics as the final metrics of this run.
        client = MlflowClient()
        runs = client.search_runs(
            [experiment_id],
            f"tags.mlflow.parentRunId = '{parent_run.info.run_id}' ",
        )
        best_nll = np.finfo(np.float64).max
        best_run = None

        nlls = [run.data.metrics["nll"] for run in runs]
        best_nll = np.min(nlls)
        best_run = runs[np.argmin(nlls)]

        mlflow.set_tag("best_run", best_run.info.run_id)
        mlflow.log_metrics(
            {
                "nll": best_nll,
            },
        )


if __name__ == "__main__":
    run()



import numpy as np
from multiprocess import Pool
from ortools.graph.python import min_cost_flow
from scipy.spatial.distance import pdist, squareform

SCALING_FACTOR = int(1e4)


def spatial_wasserstein(
    spatial_coordinates: np.ndarray,
    embeddings_truth: np.ndarray,
    embeddings_pred: np.ndarray,
    weight_scaling_factor=SCALING_FACTOR,
    demand_scaling_factor=SCALING_FACTOR,
):
    """Use max flow/min cut formulation to determine spatial Wasserstein
    distance."""

    assert len(spatial_coordinates) == len(embeddings_truth) == len(embeddings_pred)
    assert spatial_coordinates.ndim == 2
    assert embeddings_truth.ndim == embeddings_truth.ndim == 1
    assert embeddings_truth.min() >= 0
    assert embeddings_pred.min() >= 0

    if embeddings_truth.sum() == 0:
        return np.inf

    # the pairwise distances, a.k.a., weights are linearly scaled so that the minimum weight is equal to weight_scaling_factor
    pairwise_dist = squareform(pdist(spatial_coordinates))
    np.fill_diagonal(pairwise_dist, np.inf)
    pairwise_dist_min = pairwise_dist.min()
    pairwise_dist = (pairwise_dist / pairwise_dist.min() * weight_scaling_factor).astype(int)
    weight_scaling_factor_full = weight_scaling_factor / pairwise_dist_min

    # # the demands are linearly scaled so that the maximum weight is equal to demand_scaling_factor
    embeddings_truth = embeddings_truth / embeddings_truth.sum()
    embeddings_pred = embeddings_pred / embeddings_pred.sum()
    demands = embeddings_pred - embeddings_truth
    demands_max = np.abs(demands).max()
    if demands_max == 0:
        return 0
    demands = (demands / demands_max * demand_scaling_factor).astype(int)
    demand_scaling_factor_full = demand_scaling_factor / demands_max

    idx = np.argmax(np.abs(demands))
    demands[idx] -= demands.sum()

    smcf = min_cost_flow.SimpleMinCostFlow()

    edges = []

    start_nodes = []
    end_nodes = []
    capacities = []
    unit_costs = []
    for cell_idx_i, pairwise_dist_row in enumerate(pairwise_dist):
        for cell_idx_j, dist in enumerate(pairwise_dist_row):
            if cell_idx_i != cell_idx_j and demands[cell_idx_i] < 0 and demands[cell_idx_j] > 0:
                edge = (cell_idx_i, cell_idx_j, {"capacity": demand_scaling_factor, "weight": dist})
                edges.append(edge)
                start_nodes.append(cell_idx_i)
                end_nodes.append(cell_idx_j)
                capacities.append(demand_scaling_factor)
                unit_costs.append(dist)

    _ = smcf.add_arcs_with_capacity_and_unit_cost(
        start_nodes,
        end_nodes,
        capacities,
        unit_costs,
    )

    smcf.set_nodes_supplies(np.arange(0, len(demands)), -demands)

    status = smcf.solve()

    if status != smcf.OPTIMAL:
        print("There was an issue with the min cost flow input.")
        print(f"Status: {status}")
        exit(1)

    return smcf.optimal_cost() / demand_scaling_factor_full / weight_scaling_factor_full


def all_pairs_spatial_wasserstein(
    dataset,
    spatial_key: str = "spatial",
    embeddings_truth_key: str = "ground_truth_X",
    embeddings_pred_key: str = "X",
    weight_scaling_factor=SCALING_FACTOR,
    demand_scaling_factor=SCALING_FACTOR,
):
    """Compute spatial Wasserstein metric between all metagene pairs embeddings.

    Uses multiprocess.Pool to compute the distance matrix in embarassingly
    parallel fashion.

    """

    spatial_coordinates = dataset.obsm[spatial_key]
    embeddings_truth = dataset.obsm[embeddings_truth_key]
    embeddings_pred = dataset.obsm[embeddings_pred_key]

    def metric(pair):
        return spatial_wasserstein(spatial_coordinates, *pair)

    num_truth = embeddings_truth.shape[1]
    num_pred = embeddings_pred.shape[1]

    pairs = [[(truth, pred) for pred in embeddings_pred.T] for truth in embeddings_truth.T]
    pairs = np.array(pairs).reshape(-1, 2, len(spatial_coordinates))

    with Pool(processes=16) as pool:
        results = pool.map(metric, pairs)

    distances = np.array(list(results)).reshape((num_truth, num_pred))

    return distances



from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
import pymetis
from anndata import AnnData
from numpy.typing import NDArray
from pymetis import Options, part_graph
from scipy.sparse import csr_array
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import OneHotEncoder

from popari._popari_dataset import PopariDataset


class Downsampler(ABC):
    """Abstraction for downsampling spots to bins."""

    @abstractmethod
    def generate_bin_assignments(
        self,
        dataset: PopariDataset,
        bin_assignments_key: str,
        coordinates_key: str,
        downsample_rate: float,
        **bin_assignment_kwargs,
    ):
        """Generate bin assignments.

        Implementation must try to achieve the `downsample_rate` when binning, i.e. the number of
        bins should approximately equal `downsample_rate` fraction of the number of original spots.

        Returns:

        """

    @abstractmethod
    def update_binning_kwargs(previous_dataset: PopariDataset, kwargs: dict):
        """Update kwargs for the next round of bin assignment (for hierarchical
        binning).

        Since binning may need to be consistent between datasets, this allows the next round of
        binning to be informed by the results from the previous_dataset.

        Args:
            previous_dataset: the binning result from the previous round
            kwargs: keyword arguments used for the previous round of binning

        Returns:
            Updated kwargs for the next round of bin assignment.

        """

    def one_hot_encode(self, data: NDArray):
        one_hot_encoder = OneHotEncoder()
        one_hot_encoding = one_hot_encoder.fit_transform(np.array(data).reshape(-1, 1))
        one_hot_encoding = csr_array(one_hot_encoding).astype(int)

        return one_hot_encoding

    def bin_expression(self, dataset: PopariDataset, bin_assignments_key: str):

        bin_assignments = dataset.obsm[bin_assignments_key].T

        num_spots, num_genes = dataset.X.shape
        num_bins, _ = bin_assignments.shape

        binned_expression = bin_assignments @ dataset.X

        return binned_expression

    def bin_coordinates(self, dataset: PopariDataset, bin_assignments_key: str, coordinates_key: str):
        coordinates = dataset.obsm[coordinates_key]
        bin_assignments = dataset.obsm[bin_assignments_key].T

        num_assignments = np.expand_dims(bin_assignments.sum(axis=1), axis=1)
        summed_coordinates = bin_assignments @ coordinates

        binned_coordinates = summed_coordinates / num_assignments

        return binned_coordinates

    def downsample(
        self,
        dataset: PopariDataset,
        coordinates_key: str = "spatial",
        bin_assignments_key: str = "bin_assignments",
        downsample_rate: float = 0.2,
        **bin_assignment_kwargs,
    ):
        effective_kwargs = self.generate_bin_assignments(
            dataset,
            bin_assignments_key=bin_assignments_key,
            coordinates_key=coordinates_key,
            downsample_rate=downsample_rate,
            **bin_assignment_kwargs,
        )
        binned_expression = self.bin_expression(dataset, bin_assignments_key=bin_assignments_key)

        binned_dataset = AnnData(X=binned_expression)
        binned_dataset.obsm[bin_assignments_key] = dataset.obsm[bin_assignments_key].T
        binned_dataset.obsm[coordinates_key] = self.bin_coordinates(
            dataset,
            bin_assignments_key=bin_assignments_key,
            coordinates_key=coordinates_key,
        )
        binned_dataset.var_names = dataset.var_names

        for key, kwarg in effective_kwargs.items():
            binned_dataset.uns[key] = kwarg

        return binned_dataset, effective_kwargs


class GridDownsampler(Downsampler):
    """Overlay grids, and use these to downsample/bin a spatial dataset."""

    def update_binning_kwargs(previous_dataset: PopariDataset, kwargs: dict):
        for key, kwarg in kwargs.items():
            kwargs[key] = previous_dataset.uns.get(key, kwarg)

        return kwargs

    def generate_bin_assignments(
        self,
        dataset: PopariDataset,
        coordinates_key: str,
        bin_assignments_key: str,
        downsample_rate: float,
        chunks: int = 2,
        chunk_size: Optional[int] = None,
        chunk_1d_density: Optional[int] = None,
    ):
        coordinates = dataset.obsm[coordinates_key]
        binned_coordinates, chunk_size, chunk_1d_density = chunked_downsample_on_grid(
            coordinates,
            downsample_rate=downsample_rate,
            chunks=chunks,
            chunk_size=chunk_size,
            downsampled_1d_density=chunk_1d_density,
        )

        filtered_bin_coordinates = filter_gridpoints(coordinates, binned_coordinates, num_jobs=2)

        neigh = NearestNeighbors(n_neighbors=1, n_jobs=2)
        neigh.fit(filtered_bin_coordinates)
        indices = np.squeeze(neigh.kneighbors(coordinates, return_distance=False))

        bin_assignments = self.one_hot_encode(indices).T
        dataset.obsm[bin_assignments_key] = bin_assignments.T

        return {
            "chunks": None,
            "chunk_size": chunk_size,
            "chunk_1d_density": chunk_1d_density,
        }


class PartitionDownsampler(Downsampler):
    """Use graph partitioning on the Delaunay triangulation to bin a dataset."""

    def update_binning_kwargs(previous_dataset: PopariDataset, kwargs: dict):
        return kwargs

    def generate_bin_assignments(
        self,
        dataset: PopariDataset,
        adjacency_list_key: str,
        coordinates_key: str,
        bin_assignments_key: str,
        downsample_rate: float = 0.2,
    ):

        num_bins = round(len(dataset) * downsample_rate)
        adjacency_list = dataset.obsm[adjacency_list_key]

        options = Options(seed=0)  # TODO: this doesn't seem to work...
        _, indices = part_graph(num_bins, adjacency_list, options=options)

        # index_reducer = {old_index: new_index for new_index, old_index in enumerate(set(indices))}
        # reduced_indices = [index_reducer[index] for index in indices]

        bin_assignments = self.one_hot_encode(indices).T

        dataset.obsm[bin_assignments_key] = bin_assignments.T

        return {
            "adjacency_list_key": adjacency_list_key,
        }


def chunked_coordinates(coordinates: NDArray, chunks: int = None, step_size: float = None):
    """Split a list of 2D coordinates into local chunks.

    Args:
        chunks: number of equal-sized chunks to split horizontal axis. Vertical chunks are constructed
            with the same chunk size determined by splitting the horizontal axis.

    Yields:
        The next chunk of coordinates, in row-major order.

    """

    num_points, _ = coordinates.shape

    horizontal_base, vertical_base = np.min(coordinates, axis=0)
    horizontal_range, vertical_range = np.ptp(coordinates, axis=0)

    if step_size is None and chunks is None:
        raise ValueError("One of `chunks` or `step_size` must be specified.")

    if step_size is None:
        horizontal_borders, step_size = np.linspace(
            horizontal_base,
            horizontal_base + horizontal_range,
            chunks + 1,
            retstep=True,
        )
    elif chunks is None:
        horizontal_borders = np.arange(horizontal_base, horizontal_base + horizontal_range, step_size)

        # Adding endpoint
        horizontal_borders = np.append(horizontal_borders, horizontal_borders[-1] + step_size)

    vertical_borders = np.arange(vertical_base, vertical_base + vertical_range, step_size)

    # Adding endpoint
    vertical_borders = np.append(vertical_borders, vertical_borders[-1] + step_size)

    for i in range(len(horizontal_borders) - 1):
        horizontal_low, horizontal_high = horizontal_borders[i : i + 2]
        for j in range(len(vertical_borders) - 1):
            vertical_low, vertical_high = vertical_borders[j : j + 2]
            horizontal_mask = (coordinates[:, 0] > horizontal_low) & (coordinates[:, 0] <= horizontal_high)
            vertical_mask = (coordinates[:, 1] > vertical_low) & (coordinates[:, 1] <= vertical_high)
            chunk_coordinates = coordinates[horizontal_mask & vertical_mask]

            chunk_data = {
                "horizontal_low": horizontal_low,
                "horizontal_high": horizontal_high,
                "vertical_low": vertical_low,
                "vertical_high": vertical_high,
                "step_size": step_size,
                "chunk_coordinates": chunk_coordinates,
            }
            yield chunk_data


def finetune_chunk_number(coordinates: NDArray, chunks: int, downsample_rate: float, max_nudge: Optional[int] = None):
    """Heuristically search for a chunk number that cleanly splits up points.

    Using a linear search, searches for a better value of the ``chunks`` value such that
    the average points-per-chunk is closer to the number of points that will be in the chunk.

    Args:
        coordinates: original spot coordinates
        chunks: number of equal-sized chunks to split horizontal axis
        downsample_rate: approximate desired ratio of meta-spots to spots after downsampling

    Returns:
        finetuned number of chunks

    """
    if max_nudge is None:
        max_nudge = chunks // 2

    num_points, num_dimensions = coordinates.shape
    target_points = num_points * downsample_rate

    horizontal_base, vertical_base = np.min(coordinates, axis=0)
    horizontal_range, vertical_range = np.ptp(coordinates, axis=0)

    direction = 0
    for chunk_nudge in range(max_nudge):
        valid_chunks = []
        for chunk_data in chunked_coordinates(coordinates, chunks=chunks + direction * chunk_nudge):
            if len(chunk_data["chunk_coordinates"]) > 0:
                valid_chunks.append(chunk_data)

        points_per_chunk = num_points * downsample_rate / len(valid_chunks)
        downsampled_1d_density = int(np.round(np.sqrt(points_per_chunk)))
        new_direction = 1 - 2 * ((downsampled_1d_density**2) > points_per_chunk)
        if direction == 0:
            direction = new_direction
        else:
            if direction != new_direction:
                break

    return chunks + direction * chunk_nudge


def chunked_downsample_on_grid(
    coordinates: NDArray,
    downsample_rate: float,
    chunks: Optional[int] = None,
    chunk_size: Optional[float] = None,
    downsampled_1d_density: Optional[int] = None,
):
    """Downsample spot coordinates to a square grid of meta-spots using chunks.

    By chunking the coordinates, we can:

    1. Remove unused chunks.
    2. Estimate the density of spots at chunk-sized resolution.

    We use this information when downsampling in order to

    Args:
        coordinates: original spot coordinates
        chunks: number of equal-sized chunks to split horizontal axis
        downsample_rate: approximate desired ratio of meta-spots to spots after downsampling

    Returns:
        coordinates of downsampled meta-spots

    """

    num_points, num_dimensions = coordinates.shape

    horizontal_base, vertical_base = np.min(coordinates, axis=0)
    horizontal_range, vertical_range = np.ptp(coordinates, axis=0)

    if chunks is not None:
        chunks = finetune_chunk_number(coordinates, chunks, downsample_rate)

    valid_chunks = []
    for chunk_data in chunked_coordinates(coordinates, chunks=chunks, step_size=chunk_size):
        if len(chunk_data["chunk_coordinates"]) > 0:
            valid_chunks.append(chunk_data)

    points_per_chunk = num_points * downsample_rate / len(valid_chunks)

    if downsampled_1d_density is None:
        downsampled_1d_density = int(np.round(np.sqrt(points_per_chunk)))

    if points_per_chunk < 2:
        raise ValueError("Chunk density is < 1")

    all_new_coordinates = []
    for index, chunk_data in enumerate(valid_chunks):
        horizontal_low = chunk_data["horizontal_low"]
        horizontal_high = chunk_data["horizontal_high"]
        vertical_low = chunk_data["vertical_low"]
        vertical_high = chunk_data["vertical_high"]
        step_size = chunk_data["step_size"]

        x = np.linspace(horizontal_low, horizontal_high, downsampled_1d_density, endpoint=False)
        if np.allclose(horizontal_high, horizontal_base + horizontal_range):
            x_gap = x[-1] - x[-2]
            x = np.append(x, x.max() + x_gap)

        y = np.linspace(vertical_low, vertical_high, downsampled_1d_density, endpoint=False)
        if np.allclose(vertical_high, vertical_base + vertical_range):
            y_gap = y[-1] - y[-2]
            y = np.append(y, y.max() + y_gap)

        xv, yv = np.meshgrid(x, y)

        new_coordinates = np.array(list(zip(xv.flat, yv.flat)))
        all_new_coordinates.append(new_coordinates)

    new_coordinates = np.vstack(all_new_coordinates)
    new_coordinates = np.unique(new_coordinates, axis=0)

    return new_coordinates, step_size, downsampled_1d_density


def filter_gridpoints(spot_coordinates: NDArray, grid_coordinates: NDArray, num_jobs: int):
    """Use nearest neighbors approach to filter out relevant grid coordinates.

    Keeps only the grid coordinates that are mapped to at least a single original spot.

    Args:
        spot_coordinates: coordinates of original spots
        grid_coordinates: coordinates of downsampled grid

    Returns:
        metaspots that meet the filtering criterion

    """

    spot_to_metaspot_mapper = NearestNeighbors(n_neighbors=1, n_jobs=num_jobs)
    spot_to_metaspot_mapper.fit(grid_coordinates)

    indices = spot_to_metaspot_mapper.kneighbors(spot_coordinates, return_distance=False)

    used_bins = set(indices.flat)

    filtered_bin_coordinates = grid_coordinates[list(used_bins)]

    return filtered_bin_coordinates
